/*!
 * @file update.cxx
 *
 * Copyright (C) 2017 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

 //------------------------------------------------------------------------------
 // Include Files
 //------------------------------------------------------------------------------

#include <ctype.h>
#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <inttypes.h>
#include <limits.h>
#include <malloc.h>
#include <termios.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/mman.h>

#include <fastec/fi_fpga_regs.h>
#include <fastec/fi_sensor_types.h>

#include "fpga_io.h"
#include "fi_log.h"

//------------------------------------------------------------------------------
// Constant, Macros and Type Definitions
//------------------------------------------------------------------------------

// SPD EEPROM addresses
const int ISB_EEPROM_I2C_ADDR = 0x50;

// Offset in the EEPROM for the start of the sensor ID.
const int SENSOR_ID_START = 4;

const int MAX_RETRY_COUNT = 50;
const int MAX_BLOCK_COUNT = 256;
const int SECTOR_LENGTH = 65536;

const uint32_t UNLOCK_APP_SECTORS = 0x68;	//!< Block protection bits to unlock APP sectors (512-2047).
const uint32_t LOCK_SECTORS = 0x5c;		//!< Block protection bits to lock all sectors.

const uint32_t OFFSET_SECTOR = 0x7ff0000;
const uint32_t OFFSET_ADDR = 0x7fffff0;
const uint32_t APP_SECTOR = 0x2000000;

const char LINCE5M_FIRMWARE_PATH[] = "/usr/local/bin/fastec/firmware/fi_lince5m.rbf";
const char LUX13HS_FIRMWARE_PATH[] = "/usr/local/bin/fastec/firmware/fi_lux13hs.rbf";
const char LUX19HS_FIRMWARE_PATH[] = "/usr/local/bin/fastec/firmware/fi_lux19hs.rbf";

//------------------------------------------------------------------------------
// Variable Declarations
//------------------------------------------------------------------------------

bool verbose;
bool erase_only;
bool release;

char data_file[PATH_MAX];
uint8_t data_block[MAX_BLOCK_COUNT];

//------------------------------------------------------------------------------
// Prototypes
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Implementation
//------------------------------------------------------------------------------

bool i2c_wait_idle(uint8_t* data_ptr = nullptr)
{
	struct timespec sleep_req;
	int	retry_count;
	int max_retries = 100;
	uint32_t i2c_status = I2CM_STATUS_BUSY;

	// Control access to the driver.
	for (retry_count = 0; retry_count < max_retries; ++retry_count)
	{
		// Check to see if the busy bit is still set.
		if (!fpga_read_reg(I2C_MASTER_STATUS_ISB, &i2c_status) ||
			((i2c_status & I2CM_STATUS_BUSY) == 0))
			break;

		// Wait 100 milliseconds.
		sleep_req.tv_nsec = 1 * 1000 * 1000;
		sleep_req.tv_sec = 0;
		nanosleep(&sleep_req, nullptr);
	}
	if (retry_count == max_retries)
		return false;

	if (i2c_status & I2CM_STATUS_BUSY)
	{
		fi_log(LOG_DEBUG, "%s: timeout: 0x%08x", __FUNCTION__, i2c_status);
		return false;
	}
	else if ((i2c_status & I2CM_STATUS_ERR) != 0)
	{
		fi_log(LOG_DEBUG, "%s: error bit set: 0x%08x", __FUNCTION__, i2c_status);
		return false;
	}
	else if (data_ptr != nullptr)
	{
		uint32_t control;
		uint32_t data_reg;

		// Read the data.
		if (!fpga_read_reg(I2C_MASTER_RD_DATA_ISB, &data_reg))
		{
			fi_log(LOG_ERR, "Unable to read the master control register!");
			return false;
		}

		// fi_log(LOG_DEBUG, "I2C_MASTER_RD_DATA_ISB <= 0x%08x", data_reg);

		// Figure out if we have 1, 2 or 4 bytes.
		if (!fpga_read_reg(I2C_MASTER_CONTROL_ISB, &control))
		{
			fi_log(LOG_ERR, "Unable to read the master control register!");
			return false;
		}

		if ((control & I2CM_CMD_4DATA_CNT) == I2CM_CMD_4DATA_CNT)
		{
			*data_ptr = ((data_reg >> 24) & 0xFF);
			++data_ptr;
			*data_ptr = ((data_reg >> 16) & 0xFF);
			++data_ptr;
			*data_ptr = ((data_reg >> 8) & 0xFF);
			++data_ptr;
		}
		else if ((control & I2CM_CMD_2DATA_CNT) == I2CM_CMD_2DATA_CNT)
		{
			*data_ptr = ((data_reg >> 8) & 0xFF);
			++data_ptr;
		}

		*data_ptr = (data_reg & 0xFF);
	}

	// Did we exceed our retries?
	return ((i2c_status & I2CM_STATUS_BUSY) == 0);
}

bool i2c_read(uint8_t dev_addr, uint8_t reg_addr, uint8_t* reg_data, int count)
{
	bool bOK;
	uint32_t reg_value;

	// Make sure the I2C bus is available.
	if (!i2c_wait_idle())
	{
		fi_log(LOG_ERR, "%s: I2C never becomes idle.", __PRETTY_FUNCTION__);
		return false;
	}

	// Specify the register address we're going to read from.
	if (!fpga_write_reg(I2C_MASTER_CMD_DATA_ISB, (uint32_t)reg_addr))
	{
		fi_log(LOG_ERR, "%s: FPGA write error.", __PRETTY_FUNCTION__);
		return false;
	}

	// Read the data.
	reg_value = I2CM_CMD_READ;
	reg_value |= I2CM_CMD_DEV(dev_addr);
	reg_value |= I2CM_CMD_1CMD_CNT;
	if (count == 4)
		reg_value |= I2CM_CMD_4DATA_CNT;
	else if (count == 2)
		reg_value |= I2CM_CMD_2DATA_CNT;
	else if (count == 1)
		reg_value |= I2CM_CMD_1DATA_CNT;
	// fi_log(LOG_DEBUG, "I2C_MASTER_CONTROL_ISB <= 0x%08x", reg_value);
	if (!fpga_write_reg(I2C_MASTER_CONTROL_ISB, reg_value))
	{
		fi_log(LOG_ERR, "%s: FPGA write error.", __PRETTY_FUNCTION__);
		return false;
	}

	// Now wait for the data to come back.
	bOK = i2c_wait_idle(reg_data);
	if (verbose)
	{
		if (count == 4)
		{
			fi_log(LOG_DEBUG, "%s: reg_addr=0x%02X reg_data=0x%02X%02X%02X%02X", __FUNCTION__, reg_addr, *reg_data, *(reg_data + 1), *(reg_data + 2), *(reg_data + 3));
		}
		else if (count == 2)
		{
			fi_log(LOG_DEBUG, "%s: reg_addr=0x%02X reg_data=0x%02X%02X", __FUNCTION__, reg_addr, *reg_data, *(reg_data + 1));
		}
		else
		{
			fi_log(LOG_DEBUG, "%s: reg_addr=0x%02X reg_data=0x%02X", __FUNCTION__, reg_addr, *reg_data);
		}
	}

	// Return our status.
	return bOK;
}

bool i2c_write(uint8_t dev_addr, uint8_t reg_addr, uint8_t reg_data)
{
	bool bok;
	uint32_t reg_value;
	struct timespec sleep_req;

	if (verbose)
		fi_log(LOG_DEBUG, "%s: reg_addr=0x%02X reg_data=0x%02X", __FUNCTION__, reg_addr, reg_data);

	// Make sure the I2C bus is available.
	if (!i2c_wait_idle())
	{
		fi_log(LOG_ERR, "%s: I2C never becomes idle.", __PRETTY_FUNCTION__);
		return false;
	}

	// Write the data.
	if (!fpga_write_reg(I2C_MASTER_WR_DATA_ISB, (uint32_t)reg_data))
	{
		fi_log(LOG_ERR, "%s: Unable to write data word!", __PRETTY_FUNCTION__);
		return false;
	}

	// Write the register address.
	if (!fpga_write_reg(I2C_MASTER_CMD_DATA_ISB, (uint32_t)reg_addr))
	{
		fi_log(LOG_ERR, "%s: Unable to write data word!", __PRETTY_FUNCTION__);
		return false;
	}
	reg_value = I2CM_CMD_WRITE;
	reg_value |= I2CM_CMD_DEV(dev_addr);
	reg_value |= I2CM_CMD_1DATA_CNT;
	reg_value |= I2CM_CMD_1CMD_CNT;
	bok = fpga_write_reg(I2C_MASTER_CONTROL_ISB, reg_value);

	// Wait 10 milliseconds.
	sleep_req.tv_nsec = 10 * 1000 * 1000;
	sleep_req.tv_sec = 0;
	nanosleep(&sleep_req, nullptr);

	return bok;
}

static void init_defaults()
{
	verbose = false;
	erase_only = false;
	release = false;
	data_file[0] = 0;
}

static void get_release_image()
{
	int i;
	union
	{
		uint8_t rd_val[4];
		uint32_t id;
	} sensor_id;

	// Get all four bytes.
	for (i = 0; i < 4; ++i)
	{
		// See if we can read the EEPROM to figure out what sensor we're connected to.
		if (!i2c_read(ISB_EEPROM_I2C_ADDR, SENSOR_ID_START + i, &sensor_id.rd_val[i], 1))
		{
			fprintf(stderr, "Unable to read address 0x%02X\n", SENSOR_ID_START + i);
			return;
		}
	}

	// Build up the path to the firmware file.
	switch (sensor_id.id)
	{
	case FI_SENSOR_LINCE5M_MONO:
	case FI_SENSOR_LINCE5M_COLOR:
		strncpy(data_file, LINCE5M_FIRMWARE_PATH, PATH_MAX);
		break;
	case FI_SENSOR_LUX13HS_MONO:
	case FI_SENSOR_LUX13HS_COLOR:
		strncpy(data_file, LUX13HS_FIRMWARE_PATH, PATH_MAX);
		break;
	case FI_SENSOR_LUX19HS_MONO:
	case FI_SENSOR_LUX19HS_COLOR:
		strncpy(data_file, LUX19HS_FIRMWARE_PATH, PATH_MAX);
		break;
	default:
		if (verbose)
			fi_log(LOG_DEBUG, "Unknown sensor ID read [0x%08x]", sensor_id.id);
		break;
	}

	if (verbose)
		fi_log(LOG_INFO, "Update from [%s]", data_file);
}

void usage()
{
	fprintf(stderr,
		"usage: update [-r] [-v] [-e] <fpga_config_file>\n"
		" -r             --release      	  Perform a release update\n"
		"\n"
		" -v             --verbose      	  Enable verbose output\n"
		" -e             --erase        	  Only erase (no programming)\n"
	);
}

static int parse_options(int ac, char* av[])
{
	int option;
	int option_index = 0;
	static const char* short_options = "vre";
	static const struct option long_options[] =
	{
		// Flag options
		{ "help",        no_argument,       0, '?' },
		{ "verbose",     no_argument,       0, 'v' },
		{ "release",     no_argument,       0, 'r' },
		{ "erase",       no_argument,       0, 'e' },

		{ 0, 0, 0, 0 },
	};

	while (1)
	{
		option_index = 0;
		option = getopt_long(ac, av, short_options, long_options, &option_index);
		if (option == EOF)
			break;

		switch (option)
		{
			// Flags.
		case 'v':
			verbose = true;
			break;
		case 'e':
			erase_only = true;
			break;
		case 'r':
			release = true;
			break;

		default:
		case '?':
			usage();
			exit(1);
			break;
		}
	}

	if (optind < ac)
	{
		// Assuming the user specified a file.
		strncpy(data_file, av[optind], PATH_MAX);
	}

	return option_index;
}

void spinner(int pos)
{
	char spin_char;

	// We only have 4 positions.
	pos = pos % 4;

	switch (pos)
	{
	case 0:
		spin_char = '-';
		break;
	case 1:
		spin_char = '\\';
		break;
	case 2:
		spin_char = '|';
		break;
	case 3:
		spin_char = '/';
		break;
	default:
		spin_char = '+';
		break;
	}

	fprintf(stdout, "\r%c", spin_char);
	fflush(stdout);
}

bool erase_sector(uint32_t sector_addr, int count)
{
	int i;
	int sector_count;
	bool rc = false;
	uint32_t reg_value;

	if (!verbose)
		fi_log(LOG_DEBUG, "Erase 0x%08x...");

	for (sector_count = 0; sector_count < count; sector_count++)
	{
		spinner(sector_count);

		if (verbose)
			fi_log(LOG_INFO, "Erasing sector at address 0x%08x", sector_addr);

		if (!fpga_write_reg(ASMI_RD_WR_ADDRESS, sector_addr))
			goto bail;
		if (!fpga_write_reg(ASMI_COMMAND, ASMI_SECTOR_ERASE))
			goto bail;

		// Note that the datasheet says the sector erase can take up to 3 seconds. Make sure we 
		// wait around long enough!
		for (i = 0; i < MAX_RETRY_COUNT; ++i)
		{
			struct timespec sleep_req;

			if (fpga_readwait_reg(ASMI_STATUS, ASMI_ILLEGAL_ERASE_ERR | ASMI_BUSY, 0, ASMI_STATUS, &reg_value))
				break;

			// Read-Wait failed. Read the status and see why.
			if (!fpga_read_reg(ASMI_STATUS, &reg_value))
				break;
			if (verbose)
				fi_log(LOG_INFO, "STATUS = 0x%08x", reg_value);
			if (reg_value & ASMI_ILLEGAL_ERASE_ERR)
				break;

			// Wait a while
			sleep_req.tv_sec = 0;
			sleep_req.tv_nsec = 100 * 1000 * 1000;	// 100 millisecs.
			nanosleep(&sleep_req, NULL);
		}
		if (i == MAX_RETRY_COUNT)
		{
			if (verbose)
				fi_log(LOG_DEBUG, "Waiting for the sector erase took too long!");
			goto bail;
		}

		if (reg_value & ASMI_ILLEGAL_ERASE_ERR)
		{
			fi_log(LOG_ERR, "Sector erase at addr 0x%08x FAILED!", sector_addr);

			// Clear the bit.
			fpga_write_reg(ASMI_STATUS, ASMI_ILLEGAL_ERASE_ERR);
			goto bail;
		}

		// Update our sector address.
		sector_addr += 65536;
	}

	fi_log(LOG_INFO, "Erasing complete [0x%08X]", reg_value);
	rc = true;

bail:
	return rc;
}

bool flash_status()
{
	bool rc = false;
	uint32_t status;
	uint32_t reg_value;

	if (!fpga_write_reg(ASMI_COMMAND, ASMI_READ_STATUS))
		goto bail;
	if (!fpga_readwait_reg(ASMI_STATUS, ASMI_BUSY, 0, ASMI_STATUS, &reg_value))
		goto bail;

	status = (reg_value >> ASMI_READ_STATUS_SHIFT) & 0xFF;
	fi_log(LOG_INFO, "Flash status: 0x%02X [0x%08x]", status, reg_value);
	fi_log(LOG_INFO, "              %s", (status & 0x20) ? "Top" : "Bottom");
	fi_log(LOG_INFO, "              BP=0x%X", ((status & 0x40) >> 3) | ((status & 0x1E) >> 2));
	fi_log(LOG_INFO, "              WriteEnable=%d", status & 2);
	fi_log(LOG_INFO, "              WriteInProgress=%d", status & 1);

	rc = true;

bail:
	return rc;
}

bool fpga_reconfigure()
{
	if (verbose)
		fi_log(LOG_INFO, "Reconfigure starting");

	// Ask the FPGA to retrieve the status information.
	if (!fpga_write_reg(RU_COMMAND, RU_RECONFIG))
		goto bail;

	if (verbose)
		fi_log(LOG_INFO, "Reconfiguration in progress");
	return true;

bail:
	if (verbose)
		fi_log(LOG_INFO, "Failed.");
	return false;
}

bool protect_sector(uint32_t protect_code)
{
	bool rc = false;
	uint32_t reg_value;

	if (verbose)
		fi_log(LOG_INFO, "Writing protect code: 0x%X", protect_code);

	if (!fpga_write_reg(ASMI_SECTOR_PROTECT_CODE, protect_code))
		goto bail;
	if (!fpga_write_reg(ASMI_COMMAND, ASMI_SECTOR_PROTECT))
		goto bail;
	if (!fpga_readwait_reg(ASMI_STATUS, ASMI_BUSY, 0, ASMI_STATUS, &reg_value))
		goto bail;

	if (verbose)
		fi_log(LOG_INFO, "Protect complete [0x%08X]", reg_value);

	rc = true;

bail:
	return rc;
}

bool write_asmi_file(uint32_t sector_addr, int count)
{
	int i;
	bool rc = false;
	uint32_t fifo_level;
	uint32_t* p_data;
	uint32_t reg_value;
	uint32_t write_count;
	uint32_t cmd_value;
	uint32_t file_len;
	uint32_t file_pos;
	FILE* fp = nullptr;

	fi_log(LOG_INFO, "Write to %u [0x%x] from file '%s'", sector_addr, sector_addr, data_file);

	// We want to bit-reverse the file contents.
	cmd_value = ASMI_BIT_ORDER_REVERSE;

	// FIFO level should be empty.
	if (!fpga_read_reg(ASMI_WR_FIFO_LEVEL, &fifo_level))
		goto bail;
	if (fifo_level)
	{
		fi_log(LOG_ERR, "Somethings wrong, FIFO level = %d", fifo_level);
		goto bail;
	}

	// We are reading from the file and writing to the flash.
	fp = fopen(data_file, "rb");
	if (fp == NULL)
	{
		fprintf(stderr, "Unable to open '%s' [%m]\n", data_file);
		goto bail;
	}
	fseek(fp, 0, SEEK_END);
	file_len = ftell(fp);
	rewind(fp);

	while (1)
	{
		// Display some status for the user.
		file_pos = ftell(fp);
		fprintf(stdout, "\r%5.1f%%    ", ((((float)file_pos) * 100.0) / ((float)file_len)));

		// Read a block of data.
		write_count = fread(data_block, 1, MAX_BLOCK_COUNT, fp);
		if (feof(fp))
			break;
		if (ferror(fp))
		{
			fi_log(LOG_ERR, "Error reading file!");
			goto bail;
		}

		if (!fpga_write_reg(ASMI_RD_WR_ADDRESS, sector_addr))
			goto bail;
		if (!fpga_write_reg(ASMI_RD_WR_BYTE_COUNT, write_count))
			goto bail;

		// Write out the data to the FIFO.
		p_data = (uint32_t*)data_block;
		for (i = 0; i < (int)((write_count + 3) / 4); ++i)
		{
			if (!fpga_write_reg(ASMI_WR_FIFO_DATA, *p_data++))
				goto bail;
		}

		// Sanity check the FIFO level.
		if (!fpga_read_reg(ASMI_WR_FIFO_LEVEL, &fifo_level))
			goto bail;
		if (fifo_level != (uint32_t)((write_count + 3) / 4))
		{
			fi_log(LOG_ERR, "FIFO level isn't correct! (%d != %d)", fifo_level, ((write_count + 3) / 4));
			goto bail;
		}

		// Execute.
		if (!fpga_write_reg(ASMI_COMMAND, cmd_value | ASMI_WRITE))
			goto bail;

		// Did it work?
		if (!fpga_readwait_reg(ASMI_STATUS, ASMI_ILLEGAL_WRITE_ERR | ASMI_BUSY, 0, ASMI_STATUS, &reg_value))
		{
			if (!fpga_read_reg(ASMI_STATUS, &reg_value))
				goto bail;
			if (reg_value & ASMI_ILLEGAL_WRITE_ERR)
			{
				fi_log(LOG_ERR, "Write to addr 0x%08x FAILED!", sector_addr);

				// Clear the bit.
				fpga_write_reg(ASMI_STATUS, ASMI_ILLEGAL_WRITE_ERR);
			}

			// Done here.
			goto bail;
		}

		if (!fpga_read_reg(ASMI_WR_FIFO_LEVEL, &fifo_level))
			goto bail;
		if (fifo_level != 0)
			fi_log(LOG_INFO, "0x%08x: FIFO level = %d (0x%x)", sector_addr, fifo_level, fifo_level);

		// Update our address.
		sector_addr += write_count;
	}

	fi_log(LOG_INFO, "Programming complete");
	rc = true;

bail:
	if (fp)
		fclose(fp);

	return rc;
}

bool verify_asmi(uint32_t sector_addr, int count)
{
	int i;
	bool rc = false;
	uint32_t fifo_level;
	uint32_t* p_data;
	uint32_t reg_value;
	uint32_t read_count;
	uint32_t cmd_value;
	uint32_t file_len;
	uint32_t file_pos;
	FILE* fp = nullptr;

	fi_log(LOG_INFO, "Verify from %u [0x%x] with file '%s'", sector_addr, sector_addr, data_file);

	// We want to bit-reverse the file contents.
	cmd_value = ASMI_BIT_ORDER_REVERSE;

	// FIFO levels should be empty.
	if (!fpga_read_reg(ASMI_WR_FIFO_LEVEL, &fifo_level))
		goto bail;
	if (fifo_level)
	{
		fi_log(LOG_ERR, "Somethings wrong, WR_FIFO level = %d", fifo_level);
		goto bail;
	}

	if (!fpga_read_reg(ASMI_RD_FIFO_LEVEL, &fifo_level))
		goto bail;
	if (fifo_level)
	{
		fi_log(LOG_ERR, "Somethings wrong, RD_FIFO level = %d", fifo_level);
		goto bail;
	}

	// We are reading from the file and writing to the flash.
	fp = fopen(data_file, "rb");
	if (fp == NULL)
	{
		fprintf(stderr, "Unable to open '%s' [%m]\n", data_file);
		goto bail;
	}
	fseek(fp, 0, SEEK_END);
	file_len = ftell(fp);
	rewind(fp);

	while (1)
	{
		// Display some status for the user.
		file_pos = ftell(fp);
		fprintf(stdout, "\r%5.1f%%    ", ((((float)file_pos) * 100.0) / ((float)file_len)));

		// Read a block of data.
		read_count = fread(data_block, 1, MAX_BLOCK_COUNT, fp);
		if (feof(fp))
			break;
		if (ferror(fp))
		{
			fi_log(LOG_ERR, "Error reading file!");
			goto bail;
		}

		if (!fpga_write_reg(ASMI_RD_WR_ADDRESS, sector_addr))
			goto bail;
		if (!fpga_write_reg(ASMI_RD_WR_BYTE_COUNT, read_count))
			goto bail;

		// Execute a read.
		if (!fpga_write_reg(ASMI_COMMAND, cmd_value | ASMI_READ))
			goto bail;

		// Did it work?
		if (!fpga_readwait_reg(ASMI_STATUS, ASMI_BUSY, 0, ASMI_STATUS, &reg_value))
		{
			// Done here.
			fi_log(LOG_DEBUG, "fpga_readwait() failed!");
			goto bail;
		}

		// Check the amount of data in the fifo.
		if (!fpga_read_reg(ASMI_RD_FIFO_LEVEL, &fifo_level))
			goto bail;
		if (fifo_level != ((read_count + 3) / 4))
			fi_log(LOG_INFO, "0x%08x: FIFO level wrong: %d != %d", sector_addr, fifo_level, ((read_count + 3) / 4));

		// Read the data from the FIFO.
		p_data = (uint32_t*)data_block;
		for (i = 0; i < (int)((read_count + 3) / 4); ++i)
		{
			uint32_t read_val;

			if (!fpga_read_reg(ASMI_RD_FIFO_DATA, &read_val))
				goto bail;
			if (read_val != *p_data)
			{
				fi_log(LOG_DEBUG, "Verification failure [0x%x != 0x%x]", read_val, *p_data);
				goto bail;
			}

			++p_data;
		}

		// Update our address.
		sector_addr += read_count;
	}

	fi_log(LOG_INFO, "\nVerification complete");
	rc = true;

bail:
	if (fp)
		fclose(fp);

	// Make sure we reset the ASMI to clear any left-overs in the FIFO.
	// This would only be the case on a failure, but might as well clear
	// the FIFOs as a general rule.
	fpga_write_reg(ASMI_COMMAND, ASMI_RESET);
	fpga_write_reg(ASMI_COMMAND, 0);

	return rc;
}

/*!
 * @brief Bit-reverse a 32-bit value.
 * 	  From: http://aggregate.org/MAGIC/#Bit%20Reversal
 */
unsigned int reverse(register unsigned int x)
{
	x = (((x & 0xaaaaaaaa) >> 1) | ((x & 0x55555555) << 1));
	x = (((x & 0xcccccccc) >> 2) | ((x & 0x33333333) << 2));
	x = (((x & 0xf0f0f0f0) >> 4) | ((x & 0x0f0f0f0f) << 4));
	x = (((x & 0xff00ff00) >> 8) | ((x & 0x00ff00ff) << 8));

	return ((x >> 16) | (x << 16));
}

bool write_offset(uint32_t sector_addr, uint32_t app_offset)
{
	bool rc = false;
	uint32_t cmd_value;
	uint32_t fifo_level;
	uint32_t reg_value;
	uint32_t rev_offset;

	// Write the data normally.
	cmd_value = 0;

	// Get the other forms of the offset.
	rev_offset = reverse(app_offset);

	if (verbose)
	{
		fi_log(LOG_INFO, "Write 0x%08x 0x%08x 0x%08x to %u [0x%x]", app_offset, rev_offset, ~app_offset, sector_addr, sector_addr);
	}

	if (!fpga_read_reg(ASMI_WR_FIFO_LEVEL, &fifo_level))
		goto bail;
	if (fifo_level)
	{
		fi_log(LOG_ERR, "Somethings wrong, FIFO level = %d", fifo_level);
		goto bail;
	}

	if (!fpga_write_reg(ASMI_RD_WR_ADDRESS, sector_addr))
		goto bail;
	if (!fpga_write_reg(ASMI_RD_WR_BYTE_COUNT, 12))
		goto bail;

	// Write out the offset value.
	if (!fpga_write_reg(ASMI_WR_FIFO_DATA, app_offset))
		goto bail;
	if (!fpga_write_reg(ASMI_WR_FIFO_DATA, rev_offset))
		goto bail;
	if (!fpga_write_reg(ASMI_WR_FIFO_DATA, ~app_offset))
		goto bail;
	if (!fpga_read_reg(ASMI_WR_FIFO_LEVEL, &fifo_level))
		goto bail;
	if (fifo_level != 3)
	{
		fi_log(LOG_ERR, "FIFO level isn't correct! (%d != 3)", fifo_level);
		goto bail;
	}

	if (verbose)
		fi_log(LOG_DEBUG, "cmd_value | ASMI_WRITE = 0x%x", cmd_value | ASMI_WRITE);
	if (!fpga_write_reg(ASMI_COMMAND, cmd_value | ASMI_WRITE))
		goto bail;

	if (!fpga_readwait_reg(ASMI_STATUS, ASMI_ILLEGAL_WRITE_ERR | ASMI_BUSY, 0, ASMI_STATUS, &reg_value))
	{
		if (!fpga_read_reg(ASMI_STATUS, &reg_value))
			goto bail;
		if (reg_value & ASMI_ILLEGAL_WRITE_ERR)
		{
			fi_log(LOG_ERR, "Write to addr 0x%08x FAILED!", sector_addr);

			// Clear the bit.
			fpga_write_reg(ASMI_STATUS, ASMI_ILLEGAL_WRITE_ERR);
		}

		// Done here.
		goto bail;
	}

	if (verbose)
	{
		if (!fpga_read_reg(ASMI_WR_FIFO_LEVEL, &fifo_level))
			goto bail;
		fi_log(LOG_INFO, "After write, FIFO level = %d (0x%x)", fifo_level, fifo_level);
	}

	rc = true;

bail:
	return rc;
}

int main(int argc, char** argv)
{
	FILE* fp = nullptr;
	int file_sectors;
	int	option_index;
	uint32_t file_length;
	uint32_t reg_value;

	// Initialize our default values.
	init_defaults();

	// Get command line options but exit if that fails.
	option_index = parse_options(argc, argv);
	if (option_index >= argc)
		exit(1);

	// Setup the FPGA interface.
	if (!fpga_io_init())
		exit(1);

	if (release)
	{
		// Get the release information.
		get_release_image();
	}
	if (data_file[0] == 0)
	{
		usage();
		goto bail;
	}

	// First thing we want to do is enable 4-byte addressing.
	if (verbose)
		fi_log(LOG_INFO, "Enable 4-byte mode");
	if (!fpga_write_reg(ASMI_COMMAND, ASMI_EN4B_ENABLE))
		goto bail;
	if (!fpga_readwait_reg(ASMI_STATUS, ASMI_BUSY, 0, ASMI_STATUS, &reg_value))
		goto bail;

	// Read the ID/memory capacity.
	if (!fpga_write_reg(ASMI_COMMAND, ASMI_READ_RDID))
		goto bail;
	if (!fpga_readwait_reg(ASMI_STATUS, ASMI_BUSY, 0, ASMI_STATUS, &reg_value))
		goto bail;
	if (verbose)
	{
		uint32_t id = (reg_value >> ASMI_READ_RDID_SHIFT) & 0xFF;

		fi_log(LOG_INFO, "ID: 0x%02X (0x%08X)", id, reg_value);
		switch (id)
		{
		case 0x19:
			fi_log(LOG_INFO, "    EPCQ-L256");
			fi_log(LOG_INFO, "    not supported.");
			goto bail;
		case 0x20:
			fi_log(LOG_INFO, "    EPCQ-L512");
			fi_log(LOG_INFO, "    not supported.");
			goto bail;
		case 0x21:
			fi_log(LOG_INFO, "    EPCQ-L1024");
			break;
		default:
			fi_log(LOG_INFO, "    <unknown>");
			goto bail;
		}
	}

	// Find out how large the file is.
	fp = fopen(data_file, "rb");
	if (fp == NULL)
	{
		fprintf(stderr, "Unable to open '%s' [%m]\n", data_file);
		goto bail;
	}
	fseek(fp, 0L, SEEK_END);
	file_length = ftell(fp);
	fclose(fp);

	// How many sectors will the file consume?
	file_sectors = file_length / SECTOR_LENGTH;
	if ((file_length % SECTOR_LENGTH) != 0)
		++file_sectors;

	//---------------------------------------------------------------------------
	// Step 1. Unprotect the application sectors.
	if (!protect_sector(UNLOCK_APP_SECTORS))
	{
		fi_log(LOG_DEBUG, "Unable to unlock the sectors!");
		goto bail;
	}

	// Step 2. Erase the offset sector.
	if (!erase_sector(OFFSET_SECTOR, 1))
	{
		fi_log(LOG_DEBUG, "Unable to erase the offset sector!");
		goto bail;
	}

	// Step 3. Erase the application sectors.
	if (!erase_sector(APP_SECTOR, file_sectors))
	{
		fi_log(LOG_DEBUG, "Unable to erase the application sectors!");
		goto bail;
	}

	// Check to see if the user only wants to erase.
	if (erase_only)
	{
		fi_log(LOG_DEBUG, "Jumping out");
		goto do_reconfig;
	}

	// Step 4. Program the application image.
	if (!write_asmi_file(APP_SECTOR, file_sectors))
	{
		fi_log(LOG_DEBUG, "Unable to program the application image!");
		goto bail;
	}

	// Step 5. Verify the application image.
	if (!verify_asmi(APP_SECTOR, file_sectors))
	{
		fi_log(LOG_ERR, "Unable to verify the application image!");
		goto bail;
	}

	// Step 6. Program the offset sector.
	if (!write_offset(OFFSET_ADDR, APP_SECTOR))
	{
		fi_log(LOG_DEBUG, "Unable to program the offset value!");
		goto bail;
	}

	// Step 7. Protect the sectors.
	if (!protect_sector(LOCK_SECTORS))
	{
		fi_log(LOG_DEBUG, "Unable to lock the sectors!");
		goto bail;
	}

do_reconfig:
	// Step 8. Force an FPGA reconfiguration.
	fpga_reconfigure();

	if (verbose)
		fi_log(LOG_INFO, "Done");
	return 0;

bail:
	if (verbose)
		fi_log(LOG_INFO, "Bail...");
	return 0;
}

