/*!
 * @file fi_log.h
 * @brief Error and Debug Logging
 *
 * Copyright 2015 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef __FI_LOG_H__
#define __FI_LOG_H__

#include <syslog.h>

#ifdef __cplusplus
extern "C" {
#endif

//-----------------------------------------------------------------------------
// Debug macros and defines, levels can be added to.
//-----------------------------------------------------------------------------
#define PDEBUG_DEBUG        0x00000001  // Varies, whatever is being debugged
#define PDEBUG_FUNCTRACE    0x00000002  // Trace function calls.
#define PDEBUG_FPGAWRITE    0x00000010  // Trace register writes.
#define PDEBUG_FPGAREAD	    0x00000020  // Trace register writes.
#define PDEBUG_SENSOR  	    0x00000040  // Trace sensor activities.
#define PDEBUG_GIGEWRITE    0x00001000  // Trace GigE register writes.
#define PDEBUG_GIGEREAD	    0x00002000  // Trace GigE register writes.
#define PDEBUG_DOWNLOAD	    0x00004000  // Trace download activities.

// To avoid compiling in the PDEBUG messages,
// #define DISABLE_PDEBUG before including fi_log.h.
#ifdef DISABLE_PDEBUG
#define PDEBUG(level, fmt, args...) {}
#define PDEBUG_ENABLED(level) 0
#else
#undef PDEBUG             // undef it, just in case.
#define PDEBUG(level, fmt, args...) \
    do { \
	if ( level & verbosity ) { \
	    debug_log( fmt "\n", ##args ); \
	} \
    } while (0)
#define PDEBUG_ENABLED(level) ((( level ) & verbosity ) ? true : false )
#endif

// Function for PDEBUG logging.
extern void debug_log( const char *fmt, ... );

// Verbosity global variable for PDEBUG choices.
extern unsigned verbosity;

//-----------------------------------------------------------------------------
// Error logging functions.
//-----------------------------------------------------------------------------

//! Log an error that "should never happen" in a way that helps with debug.
#define LOG_UNLIKELY_ERRNO(f, l, e) fi_error_log_unlikely(f, l, e)
extern void fi_error_log_unlikely( const char *func, int line, int errno_value );

//! Normal error logging.
extern void fi_log( unsigned severity, const char *fmt, ... );

#ifdef __cplusplus
}
#endif

#endif /* !__FI_LOG_H__ */
