/*!
 * @file fpga_io.cxx
 * @brief This class provides access to FPGA registers.
 *
 * Copyright 2010,2015 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <unistd.h>
#include <string.h>
#include <stdint.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <syslog.h>
#include <stdio.h>
#include <stdlib.h>

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <fastec/fi_ioctl.h>

#include "fi_log.h"
#include "fpga_io.h"

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//!< FPGA driver file device
#define FPGA_DEV "/dev/fi_fpga"

//--------------------------------------------------------------------
// Variable declarations
//--------------------------------------------------------------------

//!< File descriptor to the FPGA driver.
static int fpga_fd = -1;

//!< Protect access to the open file fd.
static pthread_mutex_t fpga_io_mutex;

//============================================================================
// Implementation
//============================================================================

static inline void fpga_io_lock( void )
{
    pthread_mutex_lock( & fpga_io_mutex );
}

static inline void fpga_io_unlock( void )
{
    pthread_mutex_unlock( & fpga_io_mutex );
}

/*!
 * @brief One-time initialization of FPGA I/O access.
 *
 * @return true if everything went well, otherwise false.
 */
bool fpga_io_init( void )
{
    bool    retval = true;
    int	    status;
    uint32_t reg_value;

#if 0
    pthread_mutexattr_t	mta;
    // Initialize our mutex used for accessing this class.
    pthread_mutexattr_init( &mta );
    status = pthread_mutexattr_settype( &mta, PTHREAD_MUTEX_RECURSIVE );
    if ( status )
    {
	LOG_UNLIKELY_ERRNO( __PRETTY_FUNCTION__, __LINE__, status );
	retval = false;
    }
    status = pthread_mutex_init( &fpga_io_mutex, &mta );
#else
    status = pthread_mutex_init( &fpga_io_mutex, NULL );
#endif
    if ( status )
    {
	LOG_UNLIKELY_ERRNO( __PRETTY_FUNCTION__, __LINE__, status );
	retval = false;
    }

    // Open the FPGA device.
    fpga_fd = open( FPGA_DEV, O_RDWR, 0666 );
    if ( fpga_fd < 0 )
    {
	LOG_UNLIKELY_ERRNO( __PRETTY_FUNCTION__, __LINE__, errno );
	retval = false;
    }
    if ( PDEBUG_ENABLED( PDEBUG_FPGAWRITE ))
    {
	if (!fpga_read_reg( FPGA_REVISION, &reg_value ))
	{
	    LOG_UNLIKELY_ERRNO( __PRETTY_FUNCTION__, __LINE__, errno );
	    retval = false;
	}
	else
	{
	    fi_log( LOG_INFO, "FPGA Revision=0x%08x", reg_value );
	}
    }

    return retval;
}

/*!
 * @brief Make sure no one else can access the FPGA.
 */
void fpga_io_shutdown( void )
{
    fpga_io_lock();
    close( fpga_fd );
    fpga_fd = -1;
    fpga_io_unlock();
}

/*!
 * @brief Writes a value to a register in the FPGA.
 *
 * @param reg_addr	Register address we are writing to.
 * @param value		Value to write.
 *
 * @return true if everything went well, otherwise false.
 */
bool fpga_write_reg( uint32_t reg_addr, uint32_t value )
{
    FI_REG_ACCESS   fi_access;
    int	    status;
    bool    rc = true;

    // Grab the mutex.
    fpga_io_lock();

    // Initialize the structure for the write operation.
    fi_access.reg_id = reg_addr;
    fi_access.value = value;

    // Request the driver to perform the write.
    status = ioctl( fpga_fd, FI_FPGA_IOC_SETREG, &fi_access );
    if ( status < 0 )
    {
	LOG_UNLIKELY_ERRNO( __PRETTY_FUNCTION__, __LINE__, errno );
        rc = false;
    }

    // Grab the mutex.
    fpga_io_unlock();

    // See what writes we are logging.
    if ( PDEBUG_ENABLED( PDEBUG_FPGAWRITE ))
    {
	void *caller = __builtin_return_address(0);
	PDEBUG( PDEBUG_FPGAWRITE, "FPGA write_reg(%p) [0x%08x]=0x%08x", caller, reg_addr, value );
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Read the value from the specified FPGA register.
 *
 * @param reg_addr	Register address we want to read from.
 * @param value		Pointer to where we want to write the data to.
 *
 * @return true if everything went well, otherwise false.
 */
bool fpga_read_reg( uint32_t reg_addr, uint32_t *value )
{
    int		status;
    FI_REG_ACCESS	fi_access;

    // Grab the mutex.
    fpga_io_lock();

    // Initialize the structure for the read operation.
    fi_access.reg_id = reg_addr;
    fi_access.value = 0;

    // Request the driver to perform the read.
    status = ioctl( fpga_fd, FI_FPGA_IOC_GETREG, &fi_access );
    if ( status < 0 )
    {
	LOG_UNLIKELY_ERRNO( __PRETTY_FUNCTION__, __LINE__, errno );

	// Release the mutex.
	fpga_io_unlock();

        return false;
    }

    // Copy the value back and return.
    *value = fi_access.value;

    // Release the mutex.
    fpga_io_unlock();

    // See what reads we are logging.
    if ( PDEBUG_ENABLED( PDEBUG_FPGAREAD ))
    {
	void *caller = __builtin_return_address(0);
	PDEBUG( PDEBUG_FPGAREAD, "FPGA read_reg(%p) [0x%08x]=0x%08x", caller, reg_addr, *value );
    }

    return true;
}

/*!
 * @brief Wait for an "OK" status before reading the value from the specified FPGA register.
 *
 * @param status_reg	Register address of the status register.
 * @param mask		Mask for the bit(s) we are interested in.
 * @param ok_value	Value which means we are OK to proceed.
 * @param reg_addr	Register address we want to read from.
 * @param value		Pointer to where we want to write the data to.
 *
 * @return true if everything went well, otherwise false.
 */
bool fpga_readwait_reg( uint32_t status_reg, uint32_t mask, uint32_t ok_value, uint32_t reg_addr, uint32_t *value )
{
    int check;
    static int max_check = 0;
    bool rc = false;

    for (check=0; check<100; ++check)
    {
	uint32_t reg_val;
	struct timespec sleep_req;

	if ( !fpga_read_reg( status_reg, &reg_val ))
	    goto bail;
	if ((reg_val & mask) == ok_value)
	    break;

	// Wait a while
	sleep_req.tv_sec = 0;
	sleep_req.tv_nsec = 10 * 1000;	// 10 usecs.
	nanosleep( &sleep_req, NULL );
    }
    if (check == 100)
	goto bail;

    if (check > max_check)
    {
	max_check = check;
	// printf("max_check = %d/100\n", max_check);
    }

    // It's now OK to read the register.
    rc = fpga_read_reg(reg_addr, value);

bail:
    return rc;
}

/*!
 * @brief Peform a Read-Modify-Write operation on the FPGA.
 *
 * reg_value = ( reg_value & mask ) | value;
 *
 * @param reg_addr	Our FPGA register we'll be modifying.
 * @param value		Value to set into the register.
 * @param mask		Value to mask with the existing contents of the register.
 *
 * Performs
 *
 *	new_reg = ( old_reg & mask ) | value
 *
 * To set a bit
 *
 *	fpga_modify_reg( reg_name, BIT_NAME, -1 );
 *
 * To clear a bit
 *
 *	fpga_modify_reg( reg_name, 0, ~BIT_NAME );
 *
 * Example of changing a set of bits:
 *
 *	value =    ILC_BIT_DEPTH_8    | ILC_BIT_SEL_11_4;
 *	mask  = ~( ILC_BIT_DEPTH_MASK | ILC_BIT_SEL_MASK );
 *	fpga_modify_reg( reg_name, value, mask );
 *
 * @return true if everything went well, otherwise false.
 */
bool fpga_modify_reg( uint32_t reg_addr, uint32_t value, uint32_t mask )
{
    int	status;
    FI_REG_MODIFY	reg_modify;

    // Grab the mutex.
    fpga_io_lock();

    // Initialize the structure for the modify operation.
    reg_modify.reg_id = reg_addr;
    reg_modify.value = value;
    reg_modify.mask = mask;

    // Request the driver to perform the modify.
    status = ioctl( fpga_fd, FI_FPGA_IOC_MODREG, &reg_modify );
    if ( status == -1 )
    {
	LOG_UNLIKELY_ERRNO( __PRETTY_FUNCTION__, __LINE__, errno );

	// Release the mutex.
	fpga_io_unlock();
        return false;
    }

    // Release the mutex.
    fpga_io_unlock();

    // See what writes we are logging.
    if ( PDEBUG_ENABLED( PDEBUG_FPGAWRITE ))
    {
	void *caller = __builtin_return_address(0);
	PDEBUG( PDEBUG_FPGAWRITE, "FPGA modify_reg(%p) [0x%08x]=0x%08x&0x%08x", caller,
		reg_addr, value, mask );
    }

    return true;
}

/*!
 * @brief Peform a Read-Modify-Write operation to set a bit.
 *
 * @return true if everything went well, otherwise false.
 */
bool fpga_set_bit( uint32_t reg_addr, uint32_t value )
{
    int	status;
    FI_REG_MODIFY	reg_modify;

    // Grab the mutex.
    fpga_io_lock();

    // Initialize the structure for the modify operation.
    reg_modify.reg_id = reg_addr;
    reg_modify.value = value;
    reg_modify.mask = ~value;

    // Request the driver to perform the modify.
    status = ioctl( fpga_fd, FI_FPGA_IOC_MODREG, &reg_modify );
    if ( status == -1 )
    {
	LOG_UNLIKELY_ERRNO( __PRETTY_FUNCTION__, __LINE__, errno );

	// Release the mutex.
	fpga_io_unlock();
        return false;
    }

    // Release the mutex.
    fpga_io_unlock();

    // See what writes we are logging.
    if ( PDEBUG_ENABLED( PDEBUG_FPGAWRITE ))
    {
	void *caller = __builtin_return_address(0);
	PDEBUG( PDEBUG_FPGAWRITE, "FPGA set_bit(%p) [0x%08x]|=0x%08x", caller, reg_addr, value );
    }

    return true;
}

/*!
 * @brief Peform a Read-Modify-Write operation to clear a bit.
 *
 * @return true if everything went well, otherwise false.
 */
bool fpga_clear_bit( uint32_t reg_addr, uint32_t value )
{
    int	status;
    FI_REG_MODIFY	reg_modify;

    // Grab the mutex.
    fpga_io_lock();

    // Initialize the structure for the modify operation.
    reg_modify.reg_id = reg_addr;
    reg_modify.value = 0;
    reg_modify.mask = ~value;

    // Request the driver to perform the modify.
    status = ioctl( fpga_fd, FI_FPGA_IOC_MODREG, &reg_modify );
    if ( status == -1 )
    {
	LOG_UNLIKELY_ERRNO( __PRETTY_FUNCTION__, __LINE__, errno );

	// Release the mutex.
	fpga_io_unlock();
        return false;
    }

    // Release the mutex.
    fpga_io_unlock();

    // See what writes we are logging.
    if ( PDEBUG_ENABLED( PDEBUG_FPGAWRITE ))
    {
	void *caller = __builtin_return_address(0);
	PDEBUG( PDEBUG_FPGAWRITE, "FPGA clear_bit(%p) [0x%08x]&=~0x%08x", caller, reg_addr, value );
    }

    return true;
}

bool fpga_get_image(FI_IMAGE_ACCESS *im_access)
{
    bool rc = false;
    int	status;

    // Grab the mutex.
    fpga_io_lock();

    // Request the driver to perform the operation.
    status = ioctl( fpga_fd, FI_FPGA_IOC_READ_IMAGE, im_access );
    if ( status == -1 )
    {
	LOG_UNLIKELY_ERRNO( __PRETTY_FUNCTION__, __LINE__, errno );
        goto bail;
    }

    // All done.
    rc = true;

bail:
    // Release the mutex.
    fpga_io_unlock();

    return rc;
}

bool fpga_read_imfifo(uint8_t *imfifo_buffer, int byte_count)
{
    bool rc = false;
    int i;
    int	status;

    // Sanity check.
    if ((byte_count & 0xF) != 0)
    {
	fi_log(LOG_ERR, "%s: byte_count [%d] must be multiple of 16!", __PRETTY_FUNCTION__, byte_count);
	return false;
    }

    // Grab the mutex.
    fpga_io_lock();

    for (i = 0; i < byte_count; i += 16)
    {
	// Request the driver to perform the operation.
	status = ioctl( fpga_fd, FI_FPGA_IOC_READ_IM_FIFO, imfifo_buffer );
	if ( status == -1 )
	{
	    LOG_UNLIKELY_ERRNO( __PRETTY_FUNCTION__, __LINE__, errno );
	    goto bail;
	}

	imfifo_buffer += 16;
    }

    // All done.
    rc = true;

bail:
    // Release the mutex.
    fpga_io_unlock();

    return rc;
}

