/*!
 * @file dac.cxx
 * @brief Bring up the sensor board's DACs by writing to the I2C device.
 *
 * Copyright (C) 2018 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <getopt.h>
#include <errno.h>
#include <time.h>

// For non-stdio
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <fastec/fi_fpga_regs.h>

#include "fi_log.h"
#include "fpga_io.h"

//------------------------------------------------------------------------------
// Constant, Macros and Type Definitions
//------------------------------------------------------------------------------

// Our devices' address.
#define DAC_U12_I2C_ADDR		0x10
#define DAC_U14_I2C_ADDR		0x11
#define DAC_GLOBAL_I2C_ADDR		0x73

// Global address for DACs on a chip.
#define DAC_GLOBAL		0x0F

// These are the address for the U12 DAC.
#define DAC_VREF		0	/* VOUTA */
#define DAC_VLNC		1	/* VOUTB */
#define DAC_VLN			2	/* VOUTC */
#define DAC_VADH		3	/* VOUTD */
#define DAC_VOFF		4	/* VOUTE */
#define DAC_VAD4		5	/* VOUTF */
#define DAC_VLNA		6	/* VOUTG */
#define DAC_VCAS		7	/* VOUTH */

// These are the address for the U14 DAC.
#define DAC_VMUX1		0	/* VOUTA */
#define DAC_VTX2L_VTXM		1	/* VOUTB */
#define DAC_VABL_TST		2	/* VOUTC */
#define DAC_VPIX_VRSTH		3	/* VOUTD */
#define DAC_VRSTH_VRSTL		4	/* VOUTE */
#define DAC_VTXH		5	/* VOUTF */
#define DAC_VTXL		6	/* VOUTG */
#define DAC_VMUX2		7	/* VOUTH */

// Command codes.
enum COMMAND_CODES
{
    CC_WRITE_ONE = 0,		// Write to Input register n.
    CC_UPDATE_ONE,		// Update (power-up) DAC register n.
    CC_WRITE_ONE_UPDATE_ALL,	// Write to input register n, update all.
    CC_WRITE_UPDATE_ONE,	// Write to and update DAC register n.
    CC_POWERDOWN_ONE,		// Power down register n.
    CC_POWERDOWN_ALL,		// Power down chip.
    CC_SEL_INT_REF,		// Select internal reference.
    CC_SEL_EXT_REF		// Select external reference.
};

const int CC_SHIFT = 4;
const uint8_t CC_MASK = 0x0F;

typedef struct DAC_TRIPLES
{
    uint8_t chip;		// The chip address
    uint8_t reg;		// The register address.
    uint16_t val;		// The value to be written
} DAC_TRIPLES;

const int DT_SHIFT = 6;		// 10-bit values need to be left justified.

//------------------------------------------------------------------------------
// Variable Declarations
//------------------------------------------------------------------------------

//! Verbose mode option
bool verbose = false;

//! Ignore the result of i2c_wait_idle.
bool force = false;

//! Shutdown set this to true, powerup set this false.
bool down = true;

// Our table of chip, dac, value triplets.
// To make things easier to read, the values are as computed in the spreadsheet,
// but they need to be left justified before writing them out.
DAC_TRIPLES dac_settings[] = {
    {DAC_U12_I2C_ADDR, (CC_SEL_INT_REF << CC_SHIFT), 0x0000},
    {DAC_U12_I2C_ADDR, (CC_WRITE_ONE << CC_SHIFT) | DAC_VREF, 0x012C},
    {DAC_U12_I2C_ADDR, (CC_WRITE_ONE << CC_SHIFT) | DAC_VLNC, 0x00E1},
    {DAC_U12_I2C_ADDR, (CC_WRITE_ONE << CC_SHIFT) | DAC_VLN, 0x0096},
    {DAC_U12_I2C_ADDR, (CC_WRITE_ONE << CC_SHIFT) | DAC_VADH, 0x00E1},
    {DAC_U12_I2C_ADDR, (CC_WRITE_ONE << CC_SHIFT) | DAC_VOFF, 0x007D},
    {DAC_U12_I2C_ADDR, (CC_WRITE_ONE << CC_SHIFT) | DAC_VAD4, 0x0038},
    {DAC_U12_I2C_ADDR, (CC_WRITE_ONE << CC_SHIFT) | DAC_VLNA, 0x00ED},
    {DAC_U12_I2C_ADDR, (CC_WRITE_ONE << CC_SHIFT) | DAC_VCAS, 0x015E},

    {DAC_U14_I2C_ADDR, (CC_SEL_INT_REF << CC_SHIFT), 0x0000},
    {DAC_U14_I2C_ADDR, (CC_WRITE_ONE << CC_SHIFT) | DAC_VMUX1, 0x00E1},
    {DAC_U14_I2C_ADDR, (CC_WRITE_ONE << CC_SHIFT) | DAC_VTX2L_VTXM, 0x00AF},
    {DAC_U14_I2C_ADDR, (CC_WRITE_ONE << CC_SHIFT) | DAC_VABL_TST, 0x004B},
    {DAC_U14_I2C_ADDR, (CC_WRITE_ONE << CC_SHIFT) | DAC_VPIX_VRSTH, 0x01C2},
    {DAC_U14_I2C_ADDR, (CC_WRITE_ONE << CC_SHIFT) | DAC_VRSTH_VRSTL, 0x0338},
    {DAC_U14_I2C_ADDR, (CC_WRITE_ONE << CC_SHIFT) | DAC_VTXH, 0x0338},
    {DAC_U14_I2C_ADDR, (CC_WRITE_ONE << CC_SHIFT) | DAC_VTXL, 0x0000},
    {DAC_U14_I2C_ADDR, (CC_WRITE_ONE << CC_SHIFT) | DAC_VMUX2, 0x00E1},

    {DAC_GLOBAL_I2C_ADDR, (CC_UPDATE_ONE << CC_SHIFT) | DAC_GLOBAL, 0x0000},
};

//------------------------------------------------------------------------------
// Prototypes
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Implementation
//------------------------------------------------------------------------------

static void usage(void)
{
    fprintf(stderr, "usage: dac [options]\n"
            "  -d             --down      Power-down the DACs.\n"
            "  -f             --force     Force the I2C transactions.\n"
            "  -u             --up        Power-up the DACs.\n"
            "  -v             --verbose   Display debug messages.\n"
            "  -?             --help      Print usage message and exit\n");
}

bool i2c_bus_enabled()
{
    uint32_t i2c_status;

    if (!fpga_read_reg( NC1_FPGA_STATUS, &i2c_status ))
    {
	fi_log(LOG_ERR, "%s: Unable to read the status register!", __PRETTY_FUNCTION__);
	return false;
    }

    return ((i2c_status & NC1FS_I2C_ENABLED) == NC1FS_I2C_ENABLED);
}

bool i2c_wait_idle(uint8_t *data_ptr = nullptr)
{
    struct timespec sleep_req;
    int	retry_count;
    int max_retries = 100;
    uint32_t i2c_status = I2CM_STATUS_BUSY;

    // Control access to the driver.
    for (retry_count = 0; retry_count < max_retries; ++retry_count)
    {
        // Check to see if the busy bit is still set.
        if (!fpga_read_reg(I2C_MASTER_STATUS_ISB, &i2c_status) ||
            ((i2c_status & I2CM_STATUS_BUSY) == 0))
            break;

        // Wait 100 milliseconds.
        sleep_req.tv_nsec = 1 * 1000 * 1000;
        sleep_req.tv_sec = 0;
        nanosleep(&sleep_req, nullptr);
    }
    if (!force && (retry_count == max_retries))
	return false;

    if (i2c_status & I2CM_STATUS_BUSY)
    {
	fi_log(LOG_DEBUG, "%s: timeout: 0x%08x", __FUNCTION__, i2c_status);
	if (!force)
	    return false;
    }
    else if ((i2c_status & I2CM_STATUS_ERR) != 0)
    {
	fi_log(LOG_DEBUG, "%s: error bit set: 0x%08x", __FUNCTION__, i2c_status);
	if (!force)
	    return false;
    }
    else if (data_ptr != nullptr)
    {
	uint32_t control;
	uint32_t data_reg;

	// Read the data.
	if (!fpga_read_reg(I2C_MASTER_RD_DATA_ISB, &data_reg))
	{
	    fi_log(LOG_ERR, "Unable to read the master control register!");
	    return false;
	}

	// fi_log(LOG_DEBUG, "I2C_MASTER_RD_DATA_ISB <= 0x%08x", data_reg);

	// Figure out if we have 1, 2 or 4 bytes.
	if (!fpga_read_reg(I2C_MASTER_CONTROL_ISB, &control))
	{
	    fi_log(LOG_ERR, "Unable to read the master control register!");
	    return false;
	}

	if ((control & I2CM_CMD_4DATA_CNT) == I2CM_CMD_4DATA_CNT)
	{
	    *data_ptr = ((data_reg >> 24) & 0xFF);
	    ++data_ptr;
	    *data_ptr = ((data_reg >> 16) & 0xFF);
	    ++data_ptr;
	    *data_ptr = ((data_reg >> 8) & 0xFF);
	    ++data_ptr;
	}
	else if ((control & I2CM_CMD_2DATA_CNT) == I2CM_CMD_2DATA_CNT)
	{
	    *data_ptr = ((data_reg >> 8) & 0xFF);
	    ++data_ptr;
	}
	
        *data_ptr = (data_reg & 0xFF);
    }

    // Did we exceed our retries?
    if (!force)
	return ((i2c_status & I2CM_STATUS_BUSY) == 0);
    return true;
}

bool i2c_read(uint8_t dev_addr, uint8_t reg_addr, uint8_t *reg_data, int count)
{
    bool bOK;
    uint32_t reg_value;

    // Make sure the I2C bus is available.
    if (!i2c_wait_idle())
    {
        fi_log(LOG_ERR, "%s: I2C never becomes idle.", __PRETTY_FUNCTION__);
        return false;
    }

    // Specify the register address we're going to read from.
    if (!fpga_write_reg(I2C_MASTER_CMD_DATA_ISB, (uint32_t)reg_addr))
    {
        fi_log(LOG_ERR, "%s: FPGA write error.", __PRETTY_FUNCTION__);
        return false;
    }

    // Read the data.
    reg_value = I2CM_CMD_READ;
    reg_value |= I2CM_CMD_DEV(dev_addr);
    reg_value |= I2CM_CMD_1CMD_CNT;
    if (count == 4)
	reg_value |= I2CM_CMD_4DATA_CNT;
    else if (count == 2)
	reg_value |= I2CM_CMD_2DATA_CNT;
    else if (count == 1)
	reg_value |= I2CM_CMD_1DATA_CNT;
    // fi_log(LOG_DEBUG, "I2C_MASTER_CONTROL_ISB <= 0x%08x", reg_value);
    if (!fpga_write_reg(I2C_MASTER_CONTROL_ISB, reg_value))
    {
        fi_log(LOG_ERR, "%s: FPGA write error.", __PRETTY_FUNCTION__);
        return false;
    }

    // Now wait for the data to come back.
    bOK = i2c_wait_idle(reg_data);
    if (verbose)
    {
	if (count == 4)
	{
	    fi_log(LOG_DEBUG, "%s: reg_addr=0x%02X reg_data=0x%02X%02X%02X%02X", __FUNCTION__, reg_addr, *reg_data, *(reg_data+1), *(reg_data+2), *(reg_data+3));
	}
	else if (count == 2)
	{
	    fi_log(LOG_DEBUG, "%s: reg_addr=0x%02X reg_data=0x%02X%02X", __FUNCTION__, reg_addr, *reg_data, *(reg_data+1));
	}
	else
	{
	    fi_log(LOG_DEBUG, "%s: reg_addr=0x%02X reg_data=0x%02X", __FUNCTION__, reg_addr, *reg_data);
	}
    }

    // Return our status.
    return bOK;
}

bool i2c_write(uint8_t dev_addr, uint8_t reg_addr, uint16_t reg_data)
{
    bool bok;
    uint32_t reg_value;
    struct timespec sleep_req;

    if (verbose)
        fi_log(LOG_DEBUG, "%s: reg_addr=0x%02X reg_data=0x%02X", __FUNCTION__, reg_addr, reg_data);

    // Make sure the I2C bus is available.
    if (!i2c_wait_idle())
    {
        fi_log(LOG_ERR, "%s: I2C never becomes idle.", __PRETTY_FUNCTION__);
        return false;
    }

    // Write the data.
    if (!fpga_write_reg(I2C_MASTER_WR_DATA_ISB, (uint32_t)reg_data))
    {
	fi_log( LOG_ERR, "%s: Unable to write data word!", __PRETTY_FUNCTION__ );
	return false;
    }

    // Write the register address.
    if (!fpga_write_reg(I2C_MASTER_CMD_DATA_ISB, (uint32_t)reg_addr))
    {
	fi_log( LOG_ERR, "%s: Unable to write data word!", __PRETTY_FUNCTION__ );
	return false;
    }
    reg_value = I2CM_CMD_WRITE;
    reg_value |= I2CM_CMD_DEV(dev_addr);
    reg_value |= I2CM_CMD_2DATA_CNT;
    reg_value |= I2CM_CMD_1CMD_CNT;
    bok = fpga_write_reg(I2C_MASTER_CONTROL_ISB, reg_value);

    // Wait 10 milliseconds.
    sleep_req.tv_nsec = 10 * 1000 * 1000;
    sleep_req.tv_sec = 0;
    nanosleep(&sleep_req, nullptr);

    return bok;
}

static void parse_options(int ac, char *av[])
{
    int option;
    int option_index = 0;
    static const char *short_options = "dfuv";
    static const struct option long_options[] = {
    		{ "down", no_argument, 0, 'd' },
    		{ "up", no_argument, 0, 'u' },
    		{ "force", no_argument, 0, 'f' },
    		{ "verbose", no_argument, 0, 'v' },
	       	{ "help", no_argument,  0, '?'},
    		{ 0, 0, 0, 0 },
    };

    while (1)
    {
        option_index = 0;
        option = getopt_long(ac, av, short_options, long_options,
                             &option_index);
        if (option == EOF)
            break;

        switch (option)
        {
        default:
        case '?':
        case 0:
            usage();
            exit(1);
            break;
	case 'd':
	    down = true;
	    break;
	case 'f':
	    force = true;
	    break;
	case 'u':
	    down = false;
	    break;
        case 'v':
            verbose = true;
            break;
        }
    }
}

void dac_powerup()
{
    int status = -1;
    int i;

    // Loop through our table of commands.
    for (i = 0; i < (int)(sizeof(dac_settings) / sizeof(dac_settings[0])); ++i)
    {
	uint8_t chip_addr = dac_settings[i].chip; 
	uint8_t reg_addr = dac_settings[i].reg; 
	uint16_t reg_value = dac_settings[i].val << DT_SHIFT;

	if (verbose)
	    fi_log(LOG_DEBUG, "%02X[%02X] <= 0x%04X [0x%04X << %d]", chip_addr, reg_addr, reg_value, dac_settings[i].val, DT_SHIFT);

	if (!i2c_write(chip_addr, reg_addr, reg_value))
	    goto err_exit;
    }

    // All done.
    return;

err_exit:
    if ( status < 0 )
        fi_log( LOG_ERR, "%s: I/O error", __PRETTY_FUNCTION__);
}

void dac_shutdown()
{
    if (!i2c_write(DAC_GLOBAL_I2C_ADDR, (CC_POWERDOWN_ALL << CC_SHIFT) | DAC_GLOBAL, 0))
	fi_log( LOG_ERR, "%s: Unable to write shutdown command!", __PRETTY_FUNCTION__);
    else if (verbose)
	fi_log(LOG_DEBUG, "%s: Shutdown complete.", __PRETTY_FUNCTION__);
}

int main(int argc, char **argv)
{
    // Get command line options but exit if that fails.
    parse_options(argc, argv);

    // Open our driver.
    if (!fpga_io_init())
    {
	fi_log(LOG_ERR, "fpga_io_init() failed!");
	exit( 1 );
    }

    // Make sure we have access to the I2C bus.
    if (!i2c_bus_enabled())
    {
	fi_log(LOG_WARNING, "%s: I2C bus is disabled!", __PRETTY_FUNCTION__);
	goto bail;
    }

    // Are we shutting down?
    if (down)
	dac_shutdown();
    else
	dac_powerup();

bail:
    // Close our driver.
    fpga_io_shutdown();

    // All done.
    return 0;
}

