/*!
 * @file dac_poke.cxx
 * @brief Poke values to the sensor board's DACs by writing to the I2C device.
 *
 * Copyright (C) 2018 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <getopt.h>
#include <errno.h>
#include <time.h>

// For non-stdio
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <fmt/printf.h>
#include <algorithm>
#include <string>

#include <fastec/fi_fpga_regs.h>

#include "fi_log.h"
#include "fpga_io.h"

//------------------------------------------------------------------------------
// Constant, Macros and Type Definitions
//------------------------------------------------------------------------------

// Our devices' address.
#define DAC_U12_I2C_ADDR		0x10
#define DAC_U14_I2C_ADDR		0x11
#define DAC_GLOBAL_I2C_ADDR		0x73

// Global address for DACs on a chip.
#define DAC_GLOBAL		0x0F

// These are the address for the U12 DAC.
#define DAC_VREF		0	/* VOUTA */
#define DAC_VLNC		1	/* VOUTB */
#define DAC_VLN			2	/* VOUTC */
#define DAC_VADH		3	/* VOUTD */
#define DAC_VOFF		4	/* VOUTE */
#define DAC_VAD4		5	/* VOUTF */
#define DAC_VLNA		6	/* VOUTG */
#define DAC_VCAS		7	/* VOUTH */

// These are the address for the U14 DAC.
#define DAC_VMUX1		0	/* VOUTA */
#define DAC_VTX2L_VTXM		1	/* VOUTB */
#define DAC_VABL_TST		2	/* VOUTC */
#define DAC_VPIX_VRSTH		3	/* VOUTD */
#define DAC_VRSTH_VRSTL		4	/* VOUTE */
#define DAC_VTXH		5	/* VOUTF */
#define DAC_VTXL		6	/* VOUTG */
#define DAC_VMUX2		7	/* VOUTH */

// Command codes.
enum COMMAND_CODES
{
    CC_WRITE_ONE = 0,		// Write to Input register n.
    CC_UPDATE_ONE,		// Update (power-up) DAC register n.
    CC_WRITE_ONE_UPDATE_ALL,	// Write to input register n, update all.
    CC_WRITE_UPDATE_ONE,	// Write to and update DAC register n.
    CC_POWERDOWN_ONE,		// Power down register n.
    CC_POWERDOWN_ALL,		// Power down chip.
    CC_SEL_INT_REF,		// Select internal reference.
    CC_SEL_EXT_REF		// Select external reference.
};

const int CC_SHIFT = 4;
const uint8_t CC_MASK = 0x0F;

typedef struct DAC_TRIPLES
{
    uint8_t chip;		// The chip address
    uint8_t reg;		// The register address.
    const char *name;		// The register name.
} DAC_TRIPLES;

const int DT_SHIFT = 6;		// 10-bit values need to be left justified.

//------------------------------------------------------------------------------
// Variable Declarations
//------------------------------------------------------------------------------

//! Verbose mode option
bool verbose = false;

//! Ignore the result of i2c_wait_idle.
bool force = false;

//! Shutdown set this to true, powerup set this false.
bool down = true;

//! Name of the register, supplied by the user.
std::string dac_name;

//! Value to poke to the register, supplied by the user.
uint16_t dac_value;

// Our table of chip, dac, value triplets.
// To make things easier to read, the values are as computed in the spreadsheet,
// but they need to be left justified before writing them out.
DAC_TRIPLES dac_settings[] = {
    {DAC_U12_I2C_ADDR, (CC_WRITE_UPDATE_ONE << CC_SHIFT) | DAC_VREF, "VREF"},
    {DAC_U12_I2C_ADDR, (CC_WRITE_UPDATE_ONE << CC_SHIFT) | DAC_VLNC, "VLNC"},
    {DAC_U12_I2C_ADDR, (CC_WRITE_UPDATE_ONE << CC_SHIFT) | DAC_VLN, "VLN"},
    {DAC_U12_I2C_ADDR, (CC_WRITE_UPDATE_ONE << CC_SHIFT) | DAC_VADH, "VADH"},
    {DAC_U12_I2C_ADDR, (CC_WRITE_UPDATE_ONE << CC_SHIFT) | DAC_VOFF, "VOFF"},
    {DAC_U12_I2C_ADDR, (CC_WRITE_UPDATE_ONE << CC_SHIFT) | DAC_VAD4, "VAD4"},
    {DAC_U12_I2C_ADDR, (CC_WRITE_UPDATE_ONE << CC_SHIFT) | DAC_VLNA, "VLNA"},
    {DAC_U12_I2C_ADDR, (CC_WRITE_UPDATE_ONE << CC_SHIFT) | DAC_VCAS, "VCAS"},

    {DAC_U14_I2C_ADDR, (CC_WRITE_UPDATE_ONE << CC_SHIFT) | DAC_VMUX1, "VMUX1"},
    {DAC_U14_I2C_ADDR, (CC_WRITE_UPDATE_ONE << CC_SHIFT) | DAC_VTX2L_VTXM, "VTX2L"},
    {DAC_U14_I2C_ADDR, (CC_WRITE_UPDATE_ONE << CC_SHIFT) | DAC_VABL_TST, "VABL"},
    {DAC_U14_I2C_ADDR, (CC_WRITE_UPDATE_ONE << CC_SHIFT) | DAC_VPIX_VRSTH, "VPIX"},
    {DAC_U14_I2C_ADDR, (CC_WRITE_UPDATE_ONE << CC_SHIFT) | DAC_VRSTH_VRSTL, "VRSTH"},
    {DAC_U14_I2C_ADDR, (CC_WRITE_UPDATE_ONE << CC_SHIFT) | DAC_VTXH, "VTXH"},
    {DAC_U14_I2C_ADDR, (CC_WRITE_UPDATE_ONE << CC_SHIFT) | DAC_VTXL, "VTXL"},
    {DAC_U14_I2C_ADDR, (CC_WRITE_UPDATE_ONE << CC_SHIFT) | DAC_VMUX2, "VMUX2"},
};

//------------------------------------------------------------------------------
// Prototypes
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Implementation
//------------------------------------------------------------------------------

static void list()
{
    fprintf(stdout, "DAC Registers:\n"
	    "U12 chip:\n"
	    "  VREF  [0x012C]\n"
	    "  VLNC  [0x00E1]\n"
	    "  VLN   [0x0089]\n"
	    "  VADH  [0x00E1]\n"
	    "  VOFF  [0x00AF]\n"
	    "  VAD4  [0x0038]\n"
	    "  VLNA  [0x00ED]\n"
	    "  VCAS  [0x015E]\n"
	    "U14 chip:\n" 
	    "  VMUX1 [0x00E1]\n"
	    "  VTX2L [0x00AF]\n"
	    "  VABL  [0x0032]\n"
	    "  VPIX  [0x01C2]\n"
	    "  VRSTH [0x0338]\n"
	    "  VTXH  [0x0338]\n"
	    "  VTXL  [0x0000]\n"
	    "  VMUX2 [0x00E1]\n");
}

static void usage()
{
    fprintf(stdout, "usage: dac_poke [options] <reg-name> <value>\n"
            "  -f             --force     Force the I2C transactions.\n"
            "  -l             --list      Displays the list of registers.\n"
            "  -v             --verbose   Display debug messages.\n"
            "  -?             --help      Print usage message and exit\n");
    list();
}

void i2c_write(uint8_t dev_addr, uint8_t reg_addr, uint16_t reg_data)
{
    int dummy;
    uint32_t	reg_value;
    std::string cmd_str;

    // Write the data.
    cmd_str = fmt::sprintf("echo \"w 0x%0x 0x%0x\" > /proc/fi_pci/regs", I2C_MASTER_WR_DATA_ISB, reg_data);
    if (verbose)
    {
	fprintf(stderr, "%s\n", cmd_str.c_str());
    }
    dummy = system(cmd_str.c_str());

    // Write the register address.
    cmd_str = fmt::sprintf("echo \"w 0x%0x 0x%0x\" > /proc/fi_pci/regs", I2C_MASTER_CMD_DATA_ISB, reg_addr);
    if (verbose)
    {
	fprintf(stderr, "%s\n", cmd_str.c_str());
    }
    dummy = system(cmd_str.c_str());

    reg_value = I2CM_CMD_WRITE;
    reg_value |= I2CM_CMD_DEV(dev_addr);
    reg_value |= I2CM_CMD_2DATA_CNT;
    reg_value |= I2CM_CMD_1CMD_CNT;
    cmd_str = fmt::sprintf("echo \"w 0x%0x 0x%0x\" > /proc/fi_pci/regs", I2C_MASTER_CONTROL_ISB, reg_value);
    if (verbose)
    {
	fprintf(stderr, "%s\n", cmd_str.c_str());
    }
    dummy = system(cmd_str.c_str());
}

static void parse_options(int ac, char *av[])
{
    int option;
    int option_index = 0;
    static const char *short_options = "flv";
    static const struct option long_options[] = {
    		{ "force", no_argument, 0, 'f' },
    		{ "list", no_argument, 0, 'l' },
    		{ "verbose", no_argument, 0, 'v' },
	       	{ "help", no_argument,  0, '?'},
    		{ 0, 0, 0, 0 },
    };

    while (1)
    {
        option_index = 0;
        option = getopt_long(ac, av, short_options, long_options,
                             &option_index);
        if (option == EOF)
            break;

        switch (option)
        {
        default:
        case '?':
        case 0:
            usage();
            exit(1);
            break;
	case 'd':
	    down = true;
	    break;
	case 'f':
	    force = true;
	    break;
	case 'l':
	    list();
	    exit(1);
	    break;
	case 'u':
	    down = false;
	    break;
        case 'v':
            verbose = true;
            break;
        }
    }

    // Need name and value.
    if ((ac - optind) != 2)
    {
	fprintf(stderr, "Need a register name and value to poke!\n");
	usage();
	exit(1);
    }

    // Copy over the register name.
    dac_name.assign(av[optind]);
    std::transform(dac_name.begin(), dac_name.end(), dac_name.begin(), ::toupper);

    // Parse the new value.
    dac_value = (uint16_t)strtoul(av[optind + 1], NULL, 0);
    if (verbose)
	fprintf(stderr, "name='%s', value=0x%04x\n", dac_name.c_str(), dac_value);
}

int main(int argc, char **argv)
{
    int i;

    // Get command line options but exit if that fails.
    parse_options(argc, argv);

    // Loop through our table of registers, looking for the one the user
    // wants to modify.
    for (i = 0; i < (int)(sizeof(dac_settings) / sizeof(dac_settings[0])); ++i)
    {
	if (dac_name.compare(dac_settings[i].name) == 0)
	{
	    uint8_t chip_addr = dac_settings[i].chip; 
	    uint8_t reg_addr = dac_settings[i].reg; 
	    uint16_t poke_value = dac_value << DT_SHIFT;

	    if (verbose)
		fi_log(LOG_DEBUG, "%02X[%02X] <= 0x%04X [0x%04X << %d]", chip_addr, reg_addr, poke_value, dac_value, DT_SHIFT);

	    i2c_write(chip_addr, reg_addr, poke_value);
	}
    }

    // All done.
    return 0;
}

