/*!
 * @file i2c_peek.cxx
 * @brief Read values from the I2C bus.
 *
 * Copyright (C) 2018 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <getopt.h>
#include <errno.h>
#include <time.h>
#include <math.h>

// For non-stdio
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <fastec/fi_fpga_regs.h>

#include "fi_log.h"
#include "fpga_io.h"

//------------------------------------------------------------------------------
// Constant, Macros and Type Definitions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Variable Declarations
//------------------------------------------------------------------------------

//! Verbose mode option
bool verbose = false;
bool force_it = false;
bool read_one = true;
bool isb = false;
bool iob = false;
int count = 16;
uint8_t dev_addr;
uint8_t reg_addr;

//------------------------------------------------------------------------------
// Prototypes
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Implementation
//------------------------------------------------------------------------------

static void usage(void)
{
    fprintf(stderr, "usage: i2c_peek [options]\n"
	    "  -2	--two		Read two bytes (default: 1 byte)\n"
	    "  -c	--count		Number of registers to read (default: 16).\n"
	    "  -d	--dev		Device to read\n"
	    "  -f	--force		Ignore error bit.\n"
	    "  -i	--isb		Read from the ISB (default: main board)\n"
	    "  -o	--iob		Read from the IOB (default: main board)\n"
	    "  -r	--reg		Register to read\n"
            "  -v       --verbose       Make a lot of racket\n"
            "  -?       --help          Print usage message and exit\n");
}

bool i2c_wait_idle(uint8_t *data_ptr = nullptr)
{
    struct timespec sleep_req;
    int	retry_count;
    int max_retries = 10;
    uint32_t i2c_status = I2CM_STATUS_BUSY;
    uint32_t control_reg;
    uint32_t data_reg;
    uint32_t status_reg;

    if (isb)
    {
	status_reg = I2C_MASTER_STATUS_ISB;
	control_reg = I2C_MASTER_CONTROL_ISB;
	data_reg = I2C_MASTER_RD_DATA_ISB;
    }
    else if (iob)
    {
	status_reg = I2C_MASTER_STATUS_IOB;
	control_reg = I2C_MASTER_CONTROL_IOB;
	data_reg = I2C_MASTER_RD_DATA_IOB;
    }
    else
    {
	status_reg = I2C_MASTER_STATUS;
	control_reg = I2C_MASTER_CONTROL;
	data_reg = I2C_MASTER_RD_DATA;
    }

    // Control access to the driver.
    for ( retry_count = 0; retry_count < max_retries; ++retry_count )
    {
	// Check to see if the busy bit is still set.
        if (!fpga_read_reg(status_reg, &i2c_status) || ((i2c_status & I2CM_STATUS_BUSY) == 0))
	    break;

	// Wait 2 milliseconds.
	sleep_req.tv_nsec = 2 * 1000 * 1000;
	sleep_req.tv_sec = 0;
	nanosleep( &sleep_req, NULL );
    }

    if (i2c_status & I2CM_STATUS_BUSY)
    {
	if (verbose)
	    fi_log( LOG_DEBUG, "%s: timeout: 0x%08x", __FUNCTION__, i2c_status );
    }
    else if ((i2c_status & I2CM_STATUS_ERR) != 0 )
    {
	if (verbose)
	    fi_log( LOG_DEBUG, "%s: error bit set: 0x%08x", __FUNCTION__, i2c_status );
    }
    else if ( data_ptr != NULL )
    {
	uint32_t control;
	uint32_t data_val;

	// Read the data.
	if (!fpga_read_reg(data_reg, &data_val))
	{
	    fi_log(LOG_ERR, "Unable to read the master control register!");
	    return false;
	}

	// Figure out if we have 1, 2 or 4 bytes.
	if (!fpga_read_reg(control_reg, &control))
	{
	    fi_log(LOG_ERR, "Unable to read the master control register!");
	    return false;
	}

	if ((control & I2CM_CMD_4DATA_CNT) == I2CM_CMD_4DATA_CNT)
	{
	    *data_ptr = ((data_val >> 24) & 0xFF);
	    ++data_ptr;
	    *data_ptr = ((data_val >> 16) & 0xFF);
	    ++data_ptr;
	    *data_ptr = ((data_val >> 8) & 0xFF);
	    ++data_ptr;
	}
	else if ((control & I2CM_CMD_2DATA_CNT) == I2CM_CMD_2DATA_CNT)
	{
	    *data_ptr = ((data_val >> 8) & 0xFF);
	    ++data_ptr;
	}
	
        *data_ptr = (data_val & 0xFF);
    }

    // Did we exceed our retries?
    return (( i2c_status & I2CM_STATUS_BUSY ) == 0 );
}

bool i2c_read(uint8_t reg_addr, uint8_t *reg_data)
{
    bool	bOK;
    uint32_t	reg_value;
    uint32_t control_reg;
    uint32_t cmd_data_reg;

    // Make sure the I2C bus is available.
    if (!i2c_wait_idle() && !force_it)
    {
        fi_log( LOG_ERR, "%s: I2C never becomes idle.", __PRETTY_FUNCTION__ );
	return false;
    }

    if (isb)
    {
	control_reg = I2C_MASTER_CONTROL_ISB;
	cmd_data_reg = I2C_MASTER_CMD_DATA_ISB;
    }
    else if (iob)
    {
	control_reg = I2C_MASTER_CONTROL_IOB;
	cmd_data_reg = I2C_MASTER_CMD_DATA_IOB;
    }
    else
    {
	control_reg = I2C_MASTER_CONTROL;
	cmd_data_reg = I2C_MASTER_CMD_DATA;
    }

    // Specify the register address we're going to read from.
    if (!fpga_write_reg(cmd_data_reg, (uint32_t)reg_addr))
    {
        fi_log(LOG_ERR, "%s: FPGA write error.", __PRETTY_FUNCTION__);
        return false;
    }

    // Read the data, just single-byte reads.
    reg_value = I2CM_CMD_READ;
    reg_value |= I2CM_CMD_DEV(dev_addr);
    reg_value |= I2CM_CMD_1CMD_CNT;
    reg_value |= I2CM_CMD_1DATA_CNT;
    if ( !fpga_write_reg(control_reg, reg_value))
    {
        fi_log( LOG_ERR, "%s: FPGA write error.", __PRETTY_FUNCTION__ );
	return false;
    }

    // Now wait for the data to come back.
    bOK = i2c_wait_idle(reg_data);
    if (verbose)
	fi_log(LOG_DEBUG, "[%02X] = 0x%02X%02X", reg_addr, reg_data[0], reg_data[1]);

    // Return our status.
    return bOK;
}

static void parse_options(int ac, char *av[])
{
    int option;
    int option_index = 0;
    static const char *short_options = "2c:d:fior:v";
    static const struct option long_options[] = {
    		{ "count", required_argument, 0, 'c' },
    		{ "dev", required_argument, 0, 'd' },
    		{ "reg", required_argument, 0, 'r' },
    		{ "two", no_argument, 0, '2' },
    		{ "force", no_argument, 0, 'f' },
    		{ "isb", no_argument, 0, 'i' },
    		{ "iob", no_argument, 0, 'o' },
	       	{ "help", no_argument,  0, '?'},
    		{ "verbose", no_argument, 0, 'v' },
    		{ 0, 0, 0, 0 },
    };

    while (1)
    {
        option_index = 0;
        option = getopt_long(ac, av, short_options, long_options,
                             &option_index);
        if (option == EOF)
            break;

        switch (option)
        {
        default:
        case '?':
        case 0:
            usage();
            exit(1);
            break;
	case '2':
	    read_one = false;
	    break;
	case 'c':
	    count = (uint8_t)strtoul(optarg, NULL, 0);
	    break;
	case 'd':
	    dev_addr = (uint8_t)strtoul(optarg, NULL, 0);
	    break;
	case 'f':
	    force_it = true;
	    break;
	case 'i':
	    isb = true;
	    break;
	case 'o':
	    iob = true;
	    break;
	case 'r':
	    reg_addr = (uint8_t)strtoul(optarg, NULL, 0);
	    break;
        case 'v':
            verbose = true;
            break;
        }
    }

    if (verbose)
    {
	printf("bus = %c\n", (isb ? 'I' : (iob ? 'O' : 'M')));
	printf("dev_addr = 0x%02X\n", dev_addr);
	printf("reg_addr = 0x%02X\n", reg_addr);
    }
}

int main(int argc, char **argv)
{
    int i;
    uint8_t rd_val[2];
    int exit_status = 1;

    // Get command line options but exit if that fails.
    parse_options(argc, argv);

    // Open our driver.
    if ( !fpga_io_init() )
	exit( 1 );

    for (i=0; i<count; ++i)
    {
	if (!i2c_read(reg_addr + (uint8_t)i, rd_val))
	{
	    fprintf(stderr, "Unable to read address 0x%02X\n", reg_addr);
	    goto err_exit;
	}

	if ((i % 16) == 0)
	    printf("%02X: ", reg_addr + i);
	if (read_one)
	    printf("%02X ", rd_val[0]);
	else
	    printf("%02X%02x ", rd_val[0], rd_val[1]);
	if ((i % 16) == 15)
	    printf("\n");
    }
    printf("\n");

    // Return success.
    exit_status = 0;

err_exit:
    // Close our driver.
    fpga_io_shutdown();

    // All done.
    exit(exit_status);
    return exit_status;
}

