/*!
 * @file main.cxx
 * @brief The DDR SPD reader
 *
 * The purpose of this test app is to expose the contents of the DDR SPD EEPROM.
 *
 * @author Anthony L. Montiel
 * @date 2018-10-11
 *
 * Copyright (C) 2018 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

 //-----------------------------------------------------------------------------
 // Includes
 //-----------------------------------------------------------------------------

#include <ctype.h>
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <getopt.h>
#include <errno.h>
#include <time.h>

// For non-stdio
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <string>

#include <fastec/fi_fpga_regs.h>

#include "ddr4.h"
#include "fi_log.h"
#include "fpga_io.h"

//-----------------------------------------------------------------------------
// Constant Definitions
//-----------------------------------------------------------------------------

// SODIMM addresses
#define SODIMM_I2C_ADDR		0x51

#define SPD_DEFAULT_LENGTH 256
#define SPD_DDR4_LENGTH    512
#define SPD_MAX_LENGTH     SPD_DDR4_LENGTH

#define SPD_DEFAULT_PART_OFF 128
#define SPD_DDR4_PART_OFF    DDR4_SPD_REG_MODULE_PART_NUM_0

#define SPD_DEFAULT_PART_LEN 18
#define SPD_DDR4_PART_LEN    SPD_DEFAULT_PART_LEN

#define SPD_PAGE_0        (0x6C >> 1)
#define SPD_PAGE_1        (0x6E >> 1)

struct valstr
{
    int	val;
    const char *str;
};

/* different types of DRAM (fundamental memory type) for SPD */
enum spd_dram_type
{
    SPD_DRAM_TYPE_DDR = 0x07,
    SPD_DRAM_TYPE_DDR2 = 0x08,
    SPD_DRAM_TYPE_FBDIMM = 0x09,
    SPD_DRAM_TYPE_DDR3 = 0x0b,
    SPD_DRAM_TYPE_LPDDR3 = 0xf1,
    SPD_DRAM_TYPE_DDR4 = 0x0c,
    SPD_DRAM_TYPE_LPDDR4 = 0x10,
};

enum ddr3_module_type
{
    DDR3_MODULE_TYPE_UNDEFINED = 0,
    DDR3_MODULE_TYPE_RDIMM,
    DDR3_MODULE_TYPE_UDIMM,
    DDR3_MODULE_TYPE_SO_DIMM,
    DDR3_MODULE_TYPE_MICRO_DIMM,
    DDR3_MODULE_TYPE_MINI_RDIMM,
    DDR3_MODULE_TYPE_MINI_UDIMM,
    DDR3_MODULE_TYPE_MINI_CDIMM,
    DDR3_MODULE_TYPE_72b_SO_UDIMM,
    DDR3_MODULE_TYPE_72b_SO_RDIMM,
    DDR3_MODULE_TYPE_72b_SO_CDIMM,
    DDR3_MODULE_TYPE_LRDIMM,
};

const char *ddr3_module_type_lut[] = {
    [DDR3_MODULE_TYPE_UNDEFINED] = "Undefined",
    [DDR3_MODULE_TYPE_RDIMM] = "RDIMM",
    [DDR3_MODULE_TYPE_UDIMM] = "UDIMM",
    [DDR3_MODULE_TYPE_SO_DIMM] = "SO-DIMM",
    [DDR3_MODULE_TYPE_MICRO_DIMM] = "MICRO-DIMM",
    [DDR3_MODULE_TYPE_MINI_RDIMM] = "MINI-RDIMM",
    [DDR3_MODULE_TYPE_MINI_UDIMM] = "MINI-UDIMM",
    [DDR3_MODULE_TYPE_MINI_CDIMM] = "MINI-CDIMM",
    [DDR3_MODULE_TYPE_72b_SO_UDIMM] = "72b-SO-UDIMM",
    [DDR3_MODULE_TYPE_72b_SO_RDIMM] = "72b-SO-RDIMM",
    [DDR3_MODULE_TYPE_72b_SO_CDIMM] = "72b-SO-CDIMM",
    [DDR3_MODULE_TYPE_LRDIMM] = "LRDIMM",
};

enum ddr_freq
{
    DDR_FREQ_UNKNOWN = 0,	/* uninitialized value */
    DDR_333,
    DDR_400,
    DDR_533,
    DDR_667,
    DDR_800,
    DDR_933,
    DDR_1067,
    DDR_1355,
    DDR_1400,

    DDR_FREQ_MAX
};

const char *ddr_freq_prettyprint[] = {
    [DDR_FREQ_UNKNOWN] = "Unknown",
    [DDR_333] = "667",
    [DDR_400] = "800",
    [DDR_533] = "1066",
    [DDR_667] = "1333",
    [DDR_800] = "1600",
    [DDR_933] = "1866",
    [DDR_1067] = "2133",
    [DDR_1355] = "2710",
    [DDR_1400] = "2400",
};

struct spd_eeprom
{
    int length;
    uint8_t data[SPD_MAX_LENGTH];
};

/*
 * various SPD fields that can be retrieved
 * these are found in different locations on DDR/DDR2 vs. FBDIMM
 */
enum spd_field_type
{
    SPD_GET_DRAM_TYPE,      /* DRAM type */
    SPD_GET_MODULE_TYPE,    /* DIMM type */
    SPD_GET_MFG_ID,		/* Module Manufacturer ID */
    SPD_GET_MFG_ID_DRAM,	/* DRAM Manufacturer ID */
    SPD_GET_MFG_LOC,	/* Module Manufacturing Location */
    SPD_GET_MFG_DATE,	/* Module Manufacturing Date */
    SPD_GET_SERIAL_NUMBER,	/* Module Serial Number */
    SPD_GET_PART_NUMBER,	/* Module Part Number */
    SPD_GET_REVISION_CODE,	/* Module Revision Code */
    SPD_GET_SIZE,		/* Module Size (in MB) */
    SPD_GET_ECC,		/* ECC capable: boolean */
    SPD_GET_RANKS,		/* Number of ranks */
    SPD_GET_WIDTH,		/* SDRAM device width */
    SPD_GET_CHECKSUM,	/* SPD checksum */
    SPD_GET_SPEEDS,		/* module frequency capabilities */
    SPD_LAST_TYPE
};


//-----------------------------------------------------------------------------
// Variable Declarations
//-----------------------------------------------------------------------------

struct spd_eeprom spd_data;

//-----------------------------------------------------------------------------
// Our parameters which can be set at load time.
//-----------------------------------------------------------------------------

//! Verbose mode option
bool verbose = false;

//-----------------------------------------------------------------------------
// Function Prototypes
//-----------------------------------------------------------------------------

#define RETRY_COUNT 100

bool i2c_wait_idle(uint8_t *data_ptr = nullptr)
{
    struct timespec sleep_req;
    int	retry_count;
    int max_retries = 100;
    uint32_t i2c_status = I2CM_STATUS_BUSY;

    // Control access to the driver.
    for (retry_count = 0; retry_count < max_retries; ++retry_count)
    {
        // Check to see if the busy bit is still set.
        if (!fpga_read_reg(I2C_MASTER_STATUS, &i2c_status) ||
            ((i2c_status & I2CM_STATUS_BUSY) == 0))
            break;

        // Wait 100 milliseconds.
        sleep_req.tv_nsec = 1 * 1000 * 1000;
        sleep_req.tv_sec = 0;
        nanosleep(&sleep_req, nullptr);
    }
    if (retry_count == max_retries)
	return false;

    if (i2c_status & I2CM_STATUS_BUSY)
    {
	fi_log(LOG_DEBUG, "%s: timeout: 0x%08x", __FUNCTION__, i2c_status);
	return false;
    }
    else if ((i2c_status & I2CM_STATUS_ERR) != 0)
    {
	fi_log(LOG_DEBUG, "%s: error bit set: 0x%08x", __FUNCTION__, i2c_status);
	return false;
    }
    else if (data_ptr != nullptr)
    {
	uint32_t control;
	uint32_t data_reg;

	// Read the data.
	if (!fpga_read_reg(I2C_MASTER_RD_DATA, &data_reg))
	{
	    fi_log(LOG_ERR, "Unable to read the master control register!");
	    return false;
	}

	// fi_log(LOG_DEBUG, "I2C_MASTER_RD_DATA <= 0x%08x", data_reg);

	// Figure out if we have 1, 2 or 4 bytes.
	if (!fpga_read_reg(I2C_MASTER_CONTROL, &control))
	{
	    fi_log(LOG_ERR, "Unable to read the master control register!");
	    return false;
	}

	if ((control & I2CM_CMD_4DATA_CNT) == I2CM_CMD_4DATA_CNT)
	{
	    *data_ptr = ((data_reg >> 24) & 0xFF);
	    ++data_ptr;
	    *data_ptr = ((data_reg >> 16) & 0xFF);
	    ++data_ptr;
	    *data_ptr = ((data_reg >> 8) & 0xFF);
	    ++data_ptr;
	}
	else if ((control & I2CM_CMD_2DATA_CNT) == I2CM_CMD_2DATA_CNT)
	{
	    *data_ptr = ((data_reg >> 8) & 0xFF);
	    ++data_ptr;
	}
	
        *data_ptr = (data_reg & 0xFF);
    }

    // Did we exceed our retries?
    return ((i2c_status & I2CM_STATUS_BUSY) == 0);
}


bool i2c_read(uint8_t dev_addr, uint8_t reg_addr, uint8_t *reg_data, int count)
{
    bool bOK;
    uint32_t reg_value;

    // Make sure the I2C bus is available.
    if (!i2c_wait_idle())
    {
        fi_log(LOG_ERR, "%s: I2C never becomes idle.", __PRETTY_FUNCTION__);
        return false;
    }

    // Specify the register address we're going to read from.
    if (!fpga_write_reg(I2C_MASTER_CMD_DATA, (uint32_t)reg_addr))
    {
        fi_log(LOG_ERR, "%s: FPGA write error.", __PRETTY_FUNCTION__);
        return false;
    }

    // Read the data.
    reg_value = I2CM_CMD_READ;
    reg_value |= I2CM_CMD_DEV(dev_addr);
    reg_value |= I2CM_CMD_1CMD_CNT;
    if (count == 4)
	reg_value |= I2CM_CMD_4DATA_CNT;
    else if (count == 2)
	reg_value |= I2CM_CMD_2DATA_CNT;
    else if (count == 1)
	reg_value |= I2CM_CMD_1DATA_CNT;
    // fi_log(LOG_DEBUG, "I2C_MASTER_CONTROL <= 0x%08x", reg_value);
    if (!fpga_write_reg(I2C_MASTER_CONTROL, reg_value))
    {
        fi_log(LOG_ERR, "%s: FPGA write error.", __PRETTY_FUNCTION__);
        return false;
    }

    // Now wait for the data to come back.
    bOK = i2c_wait_idle(reg_data);
    if (verbose)
    {
	if (count == 4)
	{
	    fi_log(LOG_DEBUG, "%s: reg_addr=0x%02X reg_data=0x%02X%02X%02X%02X", __FUNCTION__, reg_addr, *reg_data, *(reg_data+1), *(reg_data+2), *(reg_data+3));
	}
	else if (count == 2)
	{
	    fi_log(LOG_DEBUG, "%s: reg_addr=0x%02X reg_data=0x%02X%02X", __FUNCTION__, reg_addr, *reg_data, *(reg_data+1));
	}
	else
	{
	    fi_log(LOG_DEBUG, "%s: reg_addr=0x%02X reg_data=0x%02X", __FUNCTION__, reg_addr, *reg_data);
	}
    }

    // Return our status.
    return bOK;
}

static void usage(void)
{
    fprintf(stderr, "usage: spd [options]\n"
        "  -v             --verbose        make a lot of racket\n"
        "  -?             --help           print usage message and exit\n");
}

static void parse_options(int ac, char *av[])
{
    int option;
    int option_index = 0;
    static const char *short_options = "a:v";
    static const struct option long_options[] = {
            { "verbose", no_argument, 0, 'v' },
            { "help", no_argument,  0, '?'},
            { 0, 0, 0, 0 },
    };

    while (1)
    {
        option_index = 0;
        option = getopt_long(ac, av, short_options, long_options,
            &option_index);
        if (option == EOF)
            break;

        switch (option)
        {
        default:
        case '?':
        case 0:
            usage();
            exit(1);
            break;
        case 'v':
            verbose = true;
            break;
        }
    }
}

int print_field_ddr4(const void *data, enum spd_field_type type)
{
    int ret;
    const uint8_t *byte = (const uint8_t *)data;

    ret = 0;
    switch (type)
    {
    case SPD_GET_DRAM_TYPE:
        if (byte[DDR4_SPD_REG_DEVICE_TYPE] == SPD_DRAM_TYPE_LPDDR4)
            printf("DRAM [0x%02X]: LPDDR4\n", byte[DDR4_SPD_REG_DEVICE_TYPE]);
        else
            printf("DRAM [0x%02X]: DDR4\n", byte[DDR4_SPD_REG_DEVICE_TYPE]);
        ret = 1;
        break;
    case SPD_GET_MODULE_TYPE:
        printf("MODULE: %s\n", ddr3_module_type_lut[byte[DDR4_SPD_REG_MODULE_TYPE]]);
        ret = 1;
        break;
    case SPD_GET_MFG_ID:
    {
        uint8_t manuf_lsb;
        uint8_t manuf_msb;

        manuf_lsb = byte[DDR4_SPD_REG_MODULE_MANUF_JEDEC_ID_LSB] & 0x7f;
        manuf_msb = byte[DDR4_SPD_REG_MODULE_MANUF_JEDEC_ID_MSB] & 0x7f;
	printf("MODULE_MFG: %u-%u\n", manuf_lsb + 1, manuf_msb);

        ret = 1;
        break;
    }

    case SPD_GET_MFG_ID_DRAM:
    {
        uint8_t manuf_lsb;
        uint8_t manuf_msb;

        manuf_lsb = byte[DDR4_SPD_REG_DRAM_MANUF_JEDEC_ID_LSB] & 0x7f;
        manuf_msb = byte[DDR4_SPD_REG_DRAM_MANUF_JEDEC_ID_MSB] & 0x7f;
	printf("DRAM_MFG: %u-%u\n", manuf_lsb, manuf_msb);

        ret = 1;
        break;
    }

    case SPD_GET_MFG_LOC:
        printf("MFG_LOC: 0x%02X\n", byte[DDR4_SPD_REG_MODULE_MANUF_LOC]);
        ret = 1;
        break;

    case SPD_GET_MFG_DATE: /* manufacturing date (BCD values) */
    {
        uint8_t year;
        uint8_t week;

        year = byte[DDR4_SPD_REG_MODULE_MANUF_DATE_YEAR];
        week = byte[DDR4_SPD_REG_MODULE_MANUF_DATE_YEAR];
        printf("MFG_DATE: 20%02x-wk%02x\n", week, year);
        ret = 1;
        break;
    }

    case SPD_GET_SERIAL_NUMBER:
        printf("SERIAL_NUMBER: %02X%02X%02X%02X\n",
            byte[DDR4_SPD_REG_MODULE_MANUF_SERIAL_0],
            byte[DDR4_SPD_REG_MODULE_MANUF_SERIAL_1],
            byte[DDR4_SPD_REG_MODULE_MANUF_SERIAL_2],
            byte[DDR4_SPD_REG_MODULE_MANUF_SERIAL_3]);
        ret = 1;
        break;

    case SPD_GET_PART_NUMBER:
    {
        char part[19];

        memcpy(part, &byte[DDR4_SPD_REG_MODULE_PART_NUM_0], 18);
        part[18] = '\0';
        printf("PART_NUMBER: %s\n", part);

        ret = 1;
        break;
    }

    case SPD_GET_REVISION_CODE:
        printf("REVISION_CODE: 0x%02X\n", byte[DDR4_SPD_REG_MODULE_REVISION_0]);
        ret = 1;
        break;

    case SPD_GET_SIZE:
    {
        /* See "Calculating Module Capacity" section in DDR4 SPD
         * specification for details. */
        unsigned int size;

        /* calculate the total size in MB */
        size = 256 << (byte[DDR4_SPD_REG_DENSITY_BANKS] & 0xf);
        size >>= 3; /* in terms of bytes instead of bits. */
        size *= 8 << (byte[DDR4_SPD_REG_MODULE_BUS_WIDTH] & 0x7);
        size /= 4 << (byte[DDR4_SPD_REG_MODULE_ORG] & 0x7);
        size *= 1 + ((byte[DDR4_SPD_REG_MODULE_ORG] >> 3) & 0x7);

        printf("SIZE_MB: %u\n", size);
        ret = 1;
        break;
    }

    case SPD_GET_ECC:
    {
        uint8_t bus_ext_width = byte[DDR4_SPD_REG_MODULE_BUS_WIDTH];

        bus_ext_width >>= 3;
        bus_ext_width &= 0x7;
        printf("ECC: %u\n", bus_ext_width);
        ret = 1;
        break;
    }

    case SPD_GET_RANKS:
        printf("RANKS: %d\n", 1 + ((byte[DDR4_SPD_REG_MODULE_ORG] >> 3) & 0x7));
        ret = 1;
        break;

    case SPD_GET_WIDTH:
    {
        /* Total width including ECC. */
        uint8_t width;
        width = 8 << (byte[DDR4_SPD_REG_MODULE_BUS_WIDTH] & 0x7);
        width += 8 * ((byte[DDR4_SPD_REG_MODULE_BUS_WIDTH] >> 3) & 0x7);
        printf("WIDTH: %d\n", width);
        ret = 1;
        break;
    }

    case SPD_GET_CHECKSUM:
        printf("CHECKSUM: 0x%02x%02x\n",
            byte[DDR4_SPD_REG_CRC_1],
            byte[DDR4_SPD_REG_CRC_0]);
        ret = 1;
        break;

    case SPD_GET_SPEEDS:
    {
        int i, mhz, first_entry;
        const struct valstr possible_mhz[] = {
            { 667,  "DDR4-1333" },
            { 800,  "DDR4-1600" },
            { 1200, "DDR4-2400" },
            { 0 }
        };
        int tck_mtb = byte[DDR4_SPD_REG_TCK_MIN];
        double tck_ns, mtb_ns = 0.0, ftb_ns = 0.0;
        /* fine offset is encoded in 2's complement format */
        int8_t ftb_offset = byte[DDR4_SPD_REG_FINE_OFFSET_TCK_MIN];

        mtb_ns = (double) 0.125;
        ftb_ns = (double) 0.001;
        tck_ns = tck_mtb * mtb_ns + (ftb_offset * ftb_ns);
        mhz = (int)((double)1000 / tck_ns);

        first_entry = 1;
        printf("SPEEDS [0x%02x 0x%02x %d]: ", byte[DDR4_SPD_REG_TCK_MIN], byte[DDR4_SPD_REG_FINE_OFFSET_TCK_MIN], mhz);
        for (i = 0; possible_mhz[i].val != 0; i++)
        {
            double min = possible_mhz[i].val * 0.99;

            if (min <= mhz)
            {
                if (!first_entry)
                    printf(", ");
                first_entry = 0;
                if (byte[DDR4_SPD_REG_DEVICE_TYPE] == SPD_DRAM_TYPE_LPDDR4)
                    printf("LP");
                printf("%s", possible_mhz[i].str);
            }
        }

        printf("\n");
        ret = 1;
        break;
    }

    default:
        ret = 0;	/* force "we don't handle this here */
        break;
    }

    return ret;
}

void read_spd()
{
    uint8_t *data_ptr;
    int i;

    if (verbose)
        fi_log(LOG_DEBUG, "Reading SPD...");

    data_ptr = spd_data.data;
    for (i = 0; i < SPD_MAX_LENGTH; i++, data_ptr++)
        i2c_read(SODIMM_I2C_ADDR, i, data_ptr, 1);

    if (verbose)
        fi_log(LOG_DEBUG, "...done.");
}

int main(int argc, char **argv)
{
    int i;
    int exit_status = 1;

    // Get command line options but exit if that fails.
    parse_options(argc, argv);

    // Open our driver.
    if (!fpga_io_init())
        exit(1);

    // Read the SPD data.
    read_spd();

    for (i = SPD_GET_DRAM_TYPE; i < SPD_LAST_TYPE; ++i)
        print_field_ddr4(spd_data.data, (enum spd_field_type)i);

    // Return success.
    exit_status = 0;

    // Close our driver.
    fpga_io_shutdown();

    // All done.
    exit(exit_status);
    return exit_status;
}

